#from . import operators
#from .FEMOptV3 import *
#from .FEMtools import getVFindices
#from .BVP import BVP
#from .PDEelt import PDEelt

from fc_simesh.siMesh import siMesh
from fc_simesh.siMeshElt import siMeshElt

import numpy as np
from scipy import linalg
from scipy import sparse
import scipy.sparse.linalg as splinalg
import itertools
import types
from math import *

from fc_vfemp1 import operators
from fc_vfemp1.FEMOptV3 import *
from fc_vfemp1.FEMtools import getVFindices
from fc_vfemp1.BVP import BVP,PDE

def  AssemblyP1(Th,Op,**kwargs):
  #dtype=kwargs.get('dtype',float)
  if isinstance(Th,siMeshElt):
    return AssemblyP1_siMeshElt(Th,Op,**kwargs)
    #version=kwargs.get('version', 'OptV3') 
    #if operators.isoperatorL(Op):
      #return globals()['DAssemblyP1_'+version](Th,Op,**kwargs)
    #if operators.isoperatorH(Op):
      #return globals()['HAssemblyP1_'+version](Th,Op,**kwargs)
  if isinstance(Th,siMesh):
    return AssemblyP1_siMesh(Th,Op,**kwargs)
    #bvp=BVP(Th,PDEelt(Op))
    #A,b=bvp.Assembly(Robin=False,Dirichlet=False)
    #return A
  assert False, 'Unknow Mesh %s or Operator %s '%(type(Yh),type(Op))
  
def  AssemblyP1_siMesh(Th,Op,**kwargs):
  assert isinstance(Th,siMesh)
  if operators.isoperatorL(Op):
    return AssemblyP1_siMesh_Loperator(Th,Op,**kwargs)
  if operators.isoperatorH(Op):
    return AssemblyP1_siMesh_Hoperator(Th,Op,**kwargs)
  assert False, 'Unknow operator : %s'%type(Op)
  
def AssemblyP1_siMesh_Loperator(Th,Op,**kwargs):
  assert isinstance(Th,siMesh) and operators.isoperatorL(Op)
  version=kwargs.pop('version', 'OptV3') 
  d=kwargs.pop('d',Th.d)
  all_labels=Th.sThlab[Th.find(d)]
  labels=kwargs.pop('labels',all_labels)
  dtype=kwargs.get('dtype',float)
  local=kwargs.get('local',False)
  if np.isscalar(labels):
    labels=[labels] # must be iterable
    nlab=1
  else:
    nlab=len(labels)
  if local:
    A=operators.NoneVector(nlab)
    toGlobal=operators.NoneVector(nlab)
  else:
    A=sparse.csc_matrix((Th.nq,Th.nq),dtype=dtype)
  for i in range(nlab):
    lab=labels[i]
    idx=Th.find(d,labels=lab)
    if idx is not None:
      A_lab=AssemblyP1_siMeshElt_Loperator(Th.sTh[idx],Op,**kwargs)
      if local:
        A[i]=A_lab
        toGlobal[i]=Th.sTh[idx].toGlobal
      else:
        toGlobal=Th.sTh[idx].toGlobal
        A=sum_sub_mat(A,A_lab,toGlobal,toGlobal)
  if local:
    return A,toGlobal
  else:
    return A

def AssemblyP1_siMesh_Hoperator(Th,Op,**kwargs):
  assert isinstance(Th,siMesh) and operators.isoperatorH(Op)
  #version=kwargs.pop('version', 'OptV3') 
  #d=kwargs.pop('d',Th.d)
  #all_labels=Th.sThlab[Th.find(d)
  #labels=kwargs.pop('labels',all_labels)
  dtype=kwargs.get('dtype',float)
  block=kwargs.get('block',False)
  local=kwargs.pop('local',False)
  if not block:
    local=False
  #Num=kwargs.get('Num',1)
  #if np.isscalar(labels):
    #labels=[labels] # must be iterable
    #nlab=1
  #else:
    #nlab=len(labels)
  m=Op.m
  if block:
    A=operators.NoneMatrix(m,m)
  else:
    A=sparse.csc_matrix((m*Th.nq,m*Th.nq),dtype=dtype)
    VFInd=getVFindices(Num,self.m,self.Th.nq)
  if block and local:
    for i in range(m):
      for j in range(m):
         A[i][j],toGlobal[i][j]=AssemblyP1_siMesh_Loperator(Th,Op.H[i][j],local=True,**kwargs)
    return A,toGlobal
  if block:
    for i in range(m):
      for j in range(m):
         A[i][j]=AssemblyP1_siMesh_Loperator(Th,Op.H[i][j],local=False,**kwargs)
    return A,toGlobal
  # Else not block and not local
  A=sparse.csc_matrix((m*Th.nq,m*Th.nq),dtype=dtype)
  VFInd=getVFindices(Num,self.m,self.Th.nq)
  Ind=np.arange(Th.nq)
  for i in range(m):
    I=VFInd(Ind,i)
    for j in range(m):
      J=VFInd(Ind,j)
      Aii=AssemblyP1_siMesh_Loperator(Th,Op.H[i][j],local=False,**kwargs)
      A=sum_sub_mat(A,Ai,I,J)
  return A

def  AssemblyP1_siMeshElt(sTh,Op,**kwargs):
  assert isinstance(sTh,siMeshElt)
  if operators.isoperatorL(Op):
    return AssemblyP1_siMeshElt_Loperator(sTh,Op,**kwargs)
  if operators.isoperatorH(Op):
    return AssemblyP1_siMeshElt_Hoperator(sTh,Op,**kwargs)
  assert False, 'Unknow operator : %s'%type(Op)

def  AssemblyP1_siMeshElt_Loperator(sTh,Op,**kwargs):
  #dtype=kwargs.get('dtype',float)
  assert isinstance(sTh,siMeshElt) and operators.isoperatorL(Op)
  version=kwargs.get('version', 'OptV3') 
  return globals()['DAssemblyP1_'+version](sTh,Op,**kwargs)
  
def  AssemblyP1_siMeshElt_Hoperator(sTh,Op,**kwargs):
  #dtype=kwargs.get('dtype',float)
  assert isinstance(sTh,siMeshElt) and operators.isoperatorH(Op)
  version=kwargs.get('version', 'OptV3') 
  return globals()['HAssemblyP1_'+version](sTh,Op,**kwargs)

def Apply(Lop,Th,u,**kwargs):
  # perm=lambda A: scipy.sparse.csgraph.reverse_cuthill_mckee(A)
  assert isinstance(Th,siMesh) 
  solver=kwargs.get('solver', lambda A,b: splinalg.spsolve(A,b))
  perm=kwargs.get('perm', None)
  bvp=BVP(Th,pde=PDE(Op=Lop))
  [A,b]=bvp.Assembly(local=True)
  LopM=operators.Loperator(dim=Lop.dim,d=Lop.d,a0=1)
  bvp=BVP(Th,pde=PDE(Op=LopM))
  [M,b]=bvp.Assembly(local=True)
  u=u.reshape((u.shape[0],))
  b=A*u
  if perm is not None:
    p=perm(M)
    f=np.ndarray((Th.nq,))
    f[p]=solver(M[p,::][::,p],b[p])
  else:
    f=solver(M,b)
  if isinstance(f,tuple):
    info=f[1]
    f=f[0]
  return f

def NormH1(Th,U,**kwargs):
  assert isinstance(Th,siMesh) 
  Num=kwargs.get('Num', 1)
  M=kwargs.get('Mass', None)
  isM=False
  K=kwargs.get('Stiff', None)
  isK=False
  m=len(U)//Th.nq  # division entiere
  if ( m*Th.nq != len(U) ):
    print('dimension error m=%d,nq=%d,len(U)=%d'%(m,Th.nq,len(U)))
  VFNum=getVFindices(Num,m,Th.nq)
  if M!=None:
    assert(isinstance(M,sparse.csc.csc_matrix))
    assert(M.get_shape()==(Th.nq,Th.nq))
  else:
    isM=True
    OpM=operators.Lmass(Th.dim,1.0)
    M=AssemblyP1(Th,OpM)
  if K!=None:
    assert(isinstance(K,sparse.csc.csc_matrix))
    assert(K.get_shape()==(Th.nq,Th.nq))
  else:
    isK=True
    OpS=operators.Lstiff(Th.dim)
    K=AssemblyP1(Th,OpS)
  S=0.
  I=np.arange(Th.nq)
  for i in range(m):
    UI=np.abs(U[VFNum(I,i)])
    S+=np.dot(M*UI,UI)+np.dot(K*UI,UI)
  if isM and isK:
    return np.sqrt(S),M,K
  if isM:
    return np.sqrt(S),M
  if isK:
    return np.sqrt(S),K
  return np.sqrt(S)

def NormL2(Th,U,**kwargs):
  assert isinstance(Th,siMesh) 
  Num=kwargs.get('Num', 1)
  isRetMat=False
  M=kwargs.get('Mass', None)
  if M!=None:
    assert(isinstance(M,sparse.csc.csc_matrix))
    assert(M.get_shape()==(Th.nq,Th.nq))
  else:
    isRetMat=True
    OpM=operators.Lmass(Th.dim)
    M=AssemblyP1(Th,OpM)
  m=round(U.shape[0]/Th.nq)
  assert(m*Th.nq==U.shape[0])
  VFNum=getVFindices(Num,m,Th.nq)
  S=0.
  I=np.arange(Th.nq)
  for i in range(m):
    UI=np.abs(U[VFNum(I,i)])
    S+=np.dot(M*UI,UI)
  if isRetMat:
    return np.sqrt(S),M
  else:
    return np.sqrt(S)

def Assembly_pde(pde,sTh,**kwargs):
  assert isinstance(pde,PDE)
  assert isinstance(sTh,siMeshElt)
  assert (pde.dim == sTh.dim)
  local=kwargs.get('local',True)
  Num=kwargs.get('Num',1)
  block=kwargs.get('block',False)
  #if isinstance(pde.Op,Loperator):
  if pde.m==1:
    if pde.Op is None:
      A=None
    else:
      A=FEM.DAssemblyP1_OptV3(sTh,pde.Op,local=local)
    if pde.f is not None:
      Mop=Loperator(dim=sTh.dim,d=sTh.d,a0=pde.f)
      M=FEM.DAssemblyP1_OptV3(sTh,Mop)
      f=M.sum(axis=1)
      if local:
        b=f
      else:
        b=np.zeros((sTh.nqParent,1))
        b[sTh.toParent]=f
    else:
      if local:
        b=np.zeros((sTh.nq,1))
      else:
        b=np.zeros((sTh.nqParent,1))
    if block:
      A=[A];b=[b]
    return A,b
  if isinstance(pde.Op,Hoperator):
    A=FEM.HAssemblyP1_OptV3(sTh,pde.Op,local=local,Num=Num,block=block)
    b=FEM.RHSAssembly(sTh,pde.f,m=pde.m,local=local,Num=Num,block=block)
    #print('TODO : %s line %d'%(os.path.basename(__file__),__line__))
    #print('  -> function not implemented for Hoperator')
  return A,b

def Gradient(Th,u,**kwargs):
  U=Th.eval(u)
  dim=Th.dim
  solver=kwargs.pop('solver', lambda A,b: splinalg.spsolve(A,b))
  perm=kwargs.pop('perm', None)
  # Compute Mass matrix (M)
  Lop=operators.Loperator(dim=dim,a0=1)
  M=AssemblyP1(Th,Lop)
  if perm is not None:
    p=perm(M)
    M=M[p,::][::,p]
  # Compute V=Grad(u)
  V=np.ndarray((dim,U.shape[0]))
  for i in range(dim):
    c=dim*[None];c[i]=1
    Lop=operators.Loperator(dim=dim,c=c)
    A=AssemblyP1(Th,Lop)
    b=A*U
    if perm is not None:
      b=b[p]
    V[i]=solver(M,b,**kwargs)
    if perm is not None:
      V[i,p]=V[i]    
  return V

def sum_sub_mat(A,Ai,tgr,tgc):
  if Ai is None:
    return A
  if A.shape==Ai.shape:
    return A+Ai
  (I,J,K)=sparse.find(Ai)
  A+=sparse.csc_matrix((K,(tgr[I],tgc[J])),shape=A.shape)
  return A

def csr_zero_rows(A, rows_to_zero):
  if not isinstance(A, sparse.csc_matrix):
    raise ValueError('Matrix given must be of CSR format.')
  mask=np.in1d(A.indices, rows_to_zero)
  A.data[mask]=0
  #A.eliminate_zeros()
  #A.eliminate_zeros()
  
def integrate(Th,Op,u,v):
  assert operators.isoperatorL(Op)
  U=Th.eval(u)
  V=Th.eval(v)
  A=AssemblyP1(Th,Op)
  return np.dot(A*U,V)
  



import numpy as np
import types
from scipy.sparse.linalg import spsolve

from fc_simesh.siMesh import siMesh
from fc_tools.others import func2str,is_function,is_lambda_function,is_def_function

def is_function_data(f):
  return is_function(f) or np.isscalar(f) or isinstance(f,np.ndarray) or (f is None)

def check_data(f):
  assert is_function_data(f) or isinstance(f,np.ndarray)
  if is_lambda_function(f) or is_def_function(f):
    f=np.vectorize(f)
  return f
      
def check_vector(b,dim):
  assert b is None or isinstance(b,list) or isinstance(b,np.ndarray)
  if b is None:
    return b
  if isinstance(b,list):
    assert len(b)==dim
  if isinstance(b,np.ndarray):
    assert b.shape[0]==dim
  for i in range(dim):
    b[i]=check_data(b[i])
  return b

def check_matrix(A,nl,nc):
  assert A is None or isinstance(A,list)
  if A is None:
    return A
  assert len(A)==nl
  for i in range(nl):
    A[i]=check_vector(A[i],nc)
  return A

class Loperator:
  def __init__(self,**kwargs):
    self.A=None
    self.b=None
    self.c=None
    self.a0=None
    self.dim=kwargs.get('dim', 2)
    self.d=kwargs.get('d', self.dim)
    assert(self.dim>=self.d)
    self.m=1
    #type=kwargs.get('type', None)
    fill=kwargs.get('fill', False)
    self.name=kwargs.get('name', None)
    #self.order=0;
    #if (type=='fill'):
    if fill:
      A =kwargs.get('A', NoneMatrix(self.dim,self.dim))
      b =kwargs.get('b', NoneVector(self.dim))
      c =kwargs.get('c', NoneVector(self.dim))
    else:
      A =kwargs.get('A', None)
      b =kwargs.get('b', None)
      c =kwargs.get('c', None)
    a0=kwargs.get('a0', None)
    self.A=check_matrix(A,self.dim,self.dim)
    self.b=check_vector(b,self.dim)
    self.c=check_vector(c,self.dim)
    self.a0=check_data(a0)
    
  def get_order(self):
    if self.A is not None:
      return 2
    if self.b is not None or self.c is not None :
      return 1
    return 0
      
  def __repr__(self,**kwargs):
    indent=kwargs.pop('indent','')
    full=kwargs.pop('full',True) 
    strret =  'Loperator'
    if isinstance(self.name,str):
      strret =  '[name=%s]'%(self.name)
    strret += ' : (dim,d,order) = (%d,%d,%d)\n'%(self.dim,self.d,self.get_order())
    if full:
      strret += indent+'      A : %s\n'%field2str(self.A)
      strret += indent+'      b : %s\n'%field2str(self.b)
      strret += indent+'      c : %s\n'%field2str(self.c)
      strret += indent+'     a0 : %s'%field2str(self.a0)
      return strret
    isEmpty=True
    if self.A is not None:
      isEmpty=False
      strret += indent+'      A : %s\n'%field2str(self.A)
    if self.b is not None:
      isEmpty=False
      strret += indent+'      b : %s\n'%field2str(self.b)
    if self.c is not None:
      isEmpty=False
      strret += indent+'      c : %s\n'%field2str(self.c)
    if self.a0 is not None:
      isEmpty=False
      strret += indent+'     a0 : %s\n'%field2str(self.a0)
    if isEmpty:
      strret += indent+'     Empty operator\n'
    return strret[:-1]
    
  def __str__(self,**kwargs):      
    return self.__repr__(**kwargs)
  
  def apply(self,Th,u):
    assert( isinstance(Th,siMesh) )
    assert self.A==None and self.b==None, "A and b must be None"
    dim=self.dim
    tmpOp=Loperator(dim=dim,d=self.d,c=self.c,a0=self.a0)
    U=Th.eval(u)
    bvpRHS=BVP(Th,PDEelt(tmpOp))
    A,b=bvpRHS.Assembly(local=True,Robin=False,Dirichlet=False)
    Lop=Loperator(dim=dim,d=self.d,a0=1)
    bvpLHS=BVP(Th,PDEelt(Lop))
    M,b=bvpLHS.Assembly(local=True,Robin=False,Dirichlet=False)
    res=spsolve(M,A*U) # part <Grad(u),c>+a0*u
    ## computation of div(b*u)
    #if self.b is not None:
      #for i in range(dim):
        #if self.b[i] is not None:
          #c=dim*[None];c[i]=1
          #tmpOp=Loperator(dim=dim,d=self.d,c=c)
          #bvpRHS=BVP(Th,PDEelt(tmpOp))
          #A,b=bvpRHS.Assembly(local=True,Robin=False,Dirichlet=False)
          #f=Th.eval(self.b[i])*U
          #res+=spsolve(M,A*f)
    #if self.A is not None:
      ## Compute V=Grad(u)
      #V=[]
      #for i in range(dim):
        #c=dim*[None];c[i]=1
        #tmpOp=Loperator(dim=dim,d=self.d,c=c)
        #bvpRHS=BVP(Th,PDEelt(tmpOp))
        #A,b=bvpRHS.Assembly(local=True,Robin=False,Dirichlet=False)
        #V.append(spsolve(M,A*U))
      #for i in range(dim):
        #if self.A[i] is not None:
          #c=dim*[None];c[i]=1
          #tmpOp=Loperator(dim=dim,d=self.d,c=c)
          #bvpRHS=BVP(Th,PDEelt(tmpOp))
          #A,b=bvpRHS.Assembly(local=True,Robin=False,Dirichlet=False)
          #for j in range(dim):
            #if self.A[i][j] is not None:
              #f=Th.eval(self.A[i][j])*V[j]
              #res+=spsolve(M,A*f)
    
    return res
  
  #def AssemblyP1(self,Th,**kwargs):
    #return FEM.Assembly(Th,self,**kwargs)
    
def check_H(H,m):
  assert isinstance(H,list) and len(H)==m
  for i in range(m):
    assert isinstance(H[i],list) and len(H[i])==m
    for j in range(m):
      assert isinstance(H[i][j],Loperator) or (H[i][j] is None)
  return H

class Hoperator:
  def __init__(self,**kwargs):
    self.dim=kwargs.get('dim', 2)
    self.d=kwargs.get('d', self.dim)
    assert(self.dim>=self.d)
    self.m=kwargs.get('m', 2)
    H=kwargs.get('H', NoneMatrix(self.m,self.m))
    self.name=kwargs.get('name', '')
    self.params=kwargs.get('params', None)
    #self.order=0
    self.H=check_H(H,self.m)
    
  def get_order(self):
    if self.H is None:
      return 0
    order=0
    for i in range(self.m):
      if self.H[i] is not None:
        for j in range(self.m):
          if self.H[i][j] is not None:
            assert isinstance(self.H[i][j],Loperator)
            order=max(order,self.H[i][j].get_order())
            if order==2:
              return order
    return order
  
  def set(self,i,j,Lop):
    assert( i>=0 and i<self.m)
    assert( j>=0 and j<self.m)
    assert is_Loperator(Lop) or Lop is None
    self.H[i][j]=Lop
  
  def apply(self,Th,u):
    assert( isinstance(Th,siMesh) )
    for i in range(self.m):
      for j in range(self.m):
        if self.H[i][j] is not None:
          assert self.H[i][j].A==None and self.H[i][j].b==None, "A and b must be None in all Lopetator"
    bvp=BVP(Th,pde=PDE(Op=self))
    [A,b]=bvp.Assembly()
    Lop=Loperator(dim=self.dim,a0=1)
    Hop=Hoperator(dim=self.dim,m=self.m)
    for i in range(self.m):
      Hop.H[i][i]=Lop
    bvp=BVP(Th,pde=PDE(Op=Hop))
    [M,b]=bvp.Assembly()
    if isinstance(u,list):
      U=np.concatenate(u)
    else:
      U=u
    f=spsolve(M,A*U)
    return bvp.splitsol(f)
  
  def __repr__(self,**kwargs):
    indent=kwargs.pop('indent','')
    strret =  'Hoperator %s: \n'%(self.name)
    strret += indent+'  (dim,d,m) : (%d,%d,%d)\n'%(self.dim,self.d,self.m)
    strret += indent+'      order : %d\n'%self.get_order()
    for i in range(self.m):
      for j in range(self.m):
        strret+= indent+'    H[%d][%d] : '%(i,j)
        if self.H[i][j] is None:
          strret+= 'None\n'
        else:
          strret += self.H[i][j].__str__(indent=(6+len(indent))*' ',full=False) +'\n'
    return strret[:-1]
    
  def __str__(self,*args):      
    return self.__repr__()
    
def isoperatorL(operator):
  return isinstance(operator,Loperator)
  
def isoperatorH(operator):
  return isinstance(operator,Hoperator)

def is_Loperator(operator):
  return isinstance(operator,Loperator)
  
def is_Hoperator(operator):
  return isinstance(operator,Hoperator)

def NoneVector(dim):
  V=[]
  for i in range(dim):
    V.append(None)
  return V
  
def NoneMatrix(m,n):
  M=[]
  for i in range(m):
    M.append(NoneVector(n))
  return M


# Operateurs predefinis
#######################

def Lmass(dim,**kwargs):
  a0=kwargs.get('a0',1)
  return Loperator(dim=dim,a0=a0,name='MassW')

def Hmass(dim,m,**kwargs):
  a0=kwargs.get('a0',1)
  Hop=Hoperator(dim=dim,m=m)
  if isinstance(a0,list):
    assert len(a0)==m
    for i in range(m):
      Hop.H[i][i]=Loperator(dim=dim,a0=a0[i])
  else:
    Lop=Loperator(dim=dim,a0=a0)
    for i in range(m):
      Hop.H[i][i]=Lop
  return Hop

def Lstiff(dim):
  A=NoneMatrix(dim,dim)
  for i in range(dim):
    A[i][i]=1.
  return Loperator(dim=dim,A=A,name='Stiff')
    
def StiffElasHoperator(dim,lam,mu):
  assert dim==2 or dim==3
  assert is_function(lam) or np.isscalar(lam)
  assert is_function(mu) or np.isscalar(mu)
  Hop=Hoperator(dim=dim,m=dim,name='StiffElas',params={'lam':lam,'mu':mu})
  #Hop.order=2;
  gam=None
  if np.isscalar(lam) and np.isscalar(mu):
    gam=lam+2*mu
  if is_function(lam) and np.isscalar(mu):
    if (dim==2): gam=lambda x1,x2 :lam(x1,x2) + 2*mu
    if (dim==3): gam=lambda x1,x2,x3 :lam(x1,x2,x3) + 2*mu
  if is_function(mu) and np.isscalar(lam):
    if (dim==2): gam=lambda x1,x2 :lam + 2*mu(x1,x2)
    if (dim==3): gam=lambda x1,x2,x3 :lam + 2*mu(x1,x2,x3)
  if is_function(mu) and is_function(lam):
    if (dim==2): gam=lambda x1,x2 :lam(x1,x2) + 2*mu(x1,x2)
    if (dim==3): gam=lambda x1,x2,x3 :lam(x1,x2,x3) + 2*mu(x1,x2,x3)
  assert(gam!=None)
    
  if dim==2:
    Hop.H[0][0]=Loperator(dim=dim,A=[[gam,None],[None,mu]]) 
    Hop.H[0][1]=Loperator(dim=dim,A=[[None,lam],[mu,None]]) 
    Hop.H[1][0]=Loperator(dim=dim,A=[[None,mu],[lam,None]]) 
    Hop.H[1][1]=Loperator(dim=dim,A=[[mu,None],[None,gam]])
  if dim==3:
    Hop.H[0][0]=Loperator(dim=dim,A=[[gam,None,None],[None,mu,None],[None,None,mu]]) 
    Hop.H[0][1]=Loperator(dim=dim,A=[[None,lam,None],[mu,None,None],[None,None,None]]) 
    Hop.H[0][2]=Loperator(dim=dim,A=[[None,None,lam],[None,None,None],[mu,None,None]]) 
    Hop.H[1][0]=Loperator(dim=dim,A=[[None,mu,None],[lam,None,None],[None,None,None]]) 
    Hop.H[1][1]=Loperator(dim=dim,A=[[mu,None,None],[None,gam,None],[None,None,mu]])
    Hop.H[1][2]=Loperator(dim=dim,A=[[None,None,None],[None,None,lam],[None,mu,None]])
    Hop.H[2][0]=Loperator(dim=dim,A=[[None,None,mu],[None,None,None],[lam,None,None]])
    Hop.H[2][1]=Loperator(dim=dim,A=[[None,None,None],[None,None,mu],[None,lam,None]])
    Hop.H[2][2]=Loperator(dim=dim,A=[[mu,None,None],[None,mu,None],[None,None,gam]])
  return Hop

def opfunc2str(f):
  if f is None:
    return 'None' # 
  if is_function(f):
    return 'Function'
  if np.isscalar(f):
    return 'Scalar'
  if isinstance(f,np.ndarray):
    return 'Array'
  return 'Unknow'
  
def field2str(f):
  if isinstance(f,list):
    L=[]
    for x in f:
      L.append(field2str(x))
    return str(L).replace("'","")
  return opfunc2str(f)
  
#from fc_vfemp1.PDEelt import PDEelt 
from fc_vfemp1.BVP import BVP,PDE
#from fc_vfemp1 import FEM


import  os, errno, sys
from fc_tools.others import print_packages
global fcPause

def PrintCopyright():
  print('---------------------------------------------------------------')
  print('Solving Boundary Value Problems (BVP\'s) with fc_vfemp1 package')
  print('Copyright (C) 2017 Cuvelier F.')
  print('  (LAGA/CNRS/University of Paris XIII)')
  print('---------------------------------------------------------------\n')

def run_from_ipython():
    try:
        __IPYTHON__
        return True
    except NameError:
        return False
  
def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else: raise
      
def pause(message):
  if not run_from_ipython():
    if sys.version_info[0]==3 :
      input(message)
    else:
      raw_input(message)

def print_vfemp1_packages():
  packages=['fc_tools','fc_hypermesh','fc_oogmsh','fc_simesh','fc_matplotlib4mesh','fc_simesh_matplotlib','fc_mayavi4mesh','fc_simesh_mayavi']
  print_packages(packages)


import numpy as np
from scipy import sparse
import scipy.sparse.linalg as splinalg


#from fc_simesh.siMesh import siMesh
##from fc_vfemp1 import PDEelt as Pt
#from fc_vfemp1.FEMtools import getVFindices
#from fc_vfemp1.operators import Loperator,Hoperator,check_data,is_Loperator,is_Hoperator

class PDE:
  def __init__(self,**kwargs):
    Op=kwargs.pop('Op',None)
    f=kwargs.pop('f',None)
    dim=kwargs.pop('dim',2)
    d=kwargs.pop('d',dim)
    m=kwargs.pop('m',1)
    assert len(kwargs)==0,'Unknow key arguments %s'%(", ".join([key for key in kwargs.keys()]))
      
    #delta=kwargs.get('delta',np.zeros((self.m,)))
    if Op is not None:
      assert is_Loperator(Op) or is_Hoperator(Op)
      dim=Op.dim
      m=Op.m
      d=Op.d
    else:
      if m==1:
        Op=Loperator(dim=dim,d=d)
      else:
        Op=Hoperator(dim=dim,m=m,d=d)
    self.Op=Op  
    self.dim=dim
    self.d=d # unused
    self.m=m
    self.delta=np.zeros((self.m,)) # unused (prevision for generalized boundary condition)
    self.f=None
    if f is not None:
      if m==1:
        assert is_function_data(f)
        self.f=f
      else:
        assert isinstance(f,list) and len(f)==m
        self.f=m*[None]
        for i in range(m):
          self.f[i]=check_data(f[i])
    
       
    #assert len(args)==1 or len(args)==2
    #if len(args)==1:
      #Op=args[0]
      #assert isinstance(Op,Loperator) or isinstance(Op,Hoperator) 
      #self.dtype=kwargs.get('dtype', float)
      #self.dim=Op.dim
      #self.d=Op.d
      #self.m=Op.m
      #self.Op=Op
      #self.delta=kwargs.get('delta',np.zeros((self.m,)))
      #self.f=None
      #f=kwargs.get('f',None)
      #if f is None:
        #return      
      #if isinstance(f,list):
        #assert(len(f)==self.m)
        #self.f=f
      #else:
        #self.f=[]
        #for i in range(self.m):
          #self.f.append(f)
      #return 
    #if len(args)==2:
      #dim=args[0]
      #m=args[1]
      #if m==1:
        #self.Op=Loperator(dim=dim)
      #else:
        #self.Op=Hoperator(dim=dim,m=m)

  def __str__(self,**kwargs):
    indent=kwargs.pop('indent','')    
    strret = indent+' %s object : (dim,m) = (%d,%d)\n'%(self.__class__.__name__,self.dim,self.m)
    #strret += '      d : %d\n'%self.d 
    #strret += '    dim : %d\n'%self.dim
    #strret += '      m : %d\n'%self.m
    strret += indent+'     Op : %s\n'%(self.Op.__repr__(indent=4*' '))         #self.Op.__class__.__name__
    strret += indent+'      f : '+str(self.f)+'\n'
    if isinstance(self.delta,np.ndarray):
      strret += indent+'  delta : '+np.array_str(self.delta).replace('\n','\n          ')
    else:
      strret += indent+'  delta : '+str(self.delta)
    return strret
  
  def __repr__(self,**kwargs):
    return self.__str__(**kwargs)
  
  #-> FEM.assembly_pde
  def Assembly(self,sTh,**kwargs):
    from fc_vfemp1 import FEM
    assert isinstance(sTh,siMeshElt) and self.dim == sTh.dim
    #assert( (self.dim == sTh.dim) and (self.d == sTh.d) )
    local=kwargs.get('local',True)
    Num=kwargs.get('Num',1)
    block=kwargs.get('block',False)
    #if isinstance(self.Op,Loperator):
    if self.m==1:
      if self.Op is None:
        A=None
      else:
        #A=FEM.DAssemblyP1_OptV3new(sTh,self.Op,local=local)
        A=FEM.AssemblyP1(sTh,self.Op,**kwargs)
      if self.f is not None:
        Mop=Loperator(dim=sTh.dim,d=sTh.d,a0=self.f)
        #M=FEM.DAssemblyP1_OptV3new(sTh,Mop)
        M=FEM.AssemblyP1(sTh,Mop,**kwargs)
        f=M.sum(axis=1)
        if local:
          b=f
        else:
          b=np.zeros((sTh.nqParent,1))
          b[sTh.toParent]=f
      else:
        if local:
          b=np.zeros((sTh.nq,1))
        else:
          b=np.zeros((sTh.nqParent,1))
      if block:
        A=[A];b=[b]
      return A,b
    if isinstance(self.Op,Hoperator):
      A=FEM.HAssemblyP1_OptV3(sTh,self.Op,local=local,Num=Num,block=block)
      b=FEM.RHSAssembly(sTh,self.f,m=self.m,local=local,Num=Num,block=block)
      #print('TODO : %s line %d'%(os.path.basename(__file__),__line__))
      #print('  -> function not implemented for Hoperator')
    return A,b

class BVP:
  def __init__(self,Th,**kwargs):
    assert isinstance(Th,siMesh)
    pde = kwargs.get('pde',None)
    labels = kwargs.get('labels',None)
    if pde is not None:
      assert isinstance(pde,PDE) and Th.dim == pde.dim 
      m=pde.m
    else:
      m = kwargs.get('m',1)
    self.dim=Th.dim
    self.d=Th.d
    self.Th=Th
    self.m=m
    self.ndof=self.m*Th.nq # only P1
    
    if labels is None:
      idxlab=Th.find(Th.d)
    else:
      idxlab=Th.find(Th.d,labels)
    self.pdes=[None]*Th.nsTh
    if pde is not None:
      for i in idxlab:
        self.pdes[i]=pde
    
  def __repr__(self):    
    strret = ' %s object \n'%self.__class__.__name__
    strret += '      d : %d\n'%self.d 
    strret += '    dim : %d\n'%self.dim
    strret += '      m : %d\n'%self.m
    strret += '     Th : %s\n'%self.Th.__class__.__name__
    strret += '   ndof : %d\n'%self.ndof
    #strret += '  delta : '+np.array_str(self.delta).replace('\n','\n          ')+'\n' 
    return strret
    
  def setDirichlet(self,label,fun,**kwargs):
    #comps=kwargs.get('comps',np.arange(1,self.m+1))
    comps=kwargs.get('comps',np.arange(self.m))
    idxlab=self.Th.find(self.d-1,labels=label)
    assert idxlab is not None,'labels %s not found'%str(label)
    LopD=Loperator(dim=self.dim,d=self.d-1,a0=1)
    if self.pdes[idxlab] is None:
      self.pdes[idxlab]=PDE(dim=self.dim,m=self.m,d=self.d-1)
    if self.m==1:
      self.pdes[idxlab].Op=LopD
      self.pdes[idxlab].f=check_data(fun)
      self.pdes[idxlab].delta=0
    else:
      i=0
      for cp in comps:
        # l=cp-1
        l=cp
        self.pdes[idxlab].delta[l]=0
        self.pdes[idxlab].Op.H[l][l]=LopD
        if self.pdes[idxlab].f is None:
          self.pdes[idxlab].f=self.m*[None]
        if isinstance(fun,list):
          self.pdes[idxlab].f[l]=check_data(fun[i])
        else:
          self.pdes[idxlab].f[l]=check_data(fun)
        i=i+1
  
  def setRobin(self,label,fun,**kwargs):
    #comps=kwargs.get('comps',np.arange(1,self.m+1))
    comps=kwargs.get('comps',np.arange(self.m))
    ar=kwargs.get('ar',None) # By default Neumann
    idxlab=self.Th.find(self.d-1,labels=label)
    if idxlab is None:
      print('Error with setRobin: ...')
      return
    if ar is None:
      LopD=None
    else:
      LopD=Loperator(dim=self.dim,d=self.d-1,a0=ar)
      
    if self.pdes[idxlab] is None:
      self.pdes[idxlab]=PDE(dim=self.dim,m=self.m,d=self.d-1)
    if self.m==1:
      self.pdes[idxlab].Op=LopD
      self.pdes[idxlab].f=check_data(fun)
      self.pdes[idxlab].delta=1
    else:
      i=0
      for cp in comps:
        #l=cp-1
        l=cp
        self.pdes[idxlab].delta[l]=1
        self.pdes[idxlab].Op.H[l][l]=LopD
        if self.pdes[idxlab].f is None:
          self.pdes[idxlab].f=self.m*[None]
        if isinstance(fun,list):
          self.pdes[idxlab].f[l]=check_data(fun[i])
        else:
          self.pdes[idxlab].f[l]=check_data(fun)
        i=i+1
  
  def setPDE(self,pde,**kwargs):
    d=kwargs.pop('d',self.d)
    assert self.dim==pde.dim and self.m == pde.m
    labels=kwargs.pop('labels',self.Th.find(d))
    idxlab=self.Th.find(d,labels=labels)
    assert idxlab is not None,'mesh label(s) not found'
    if np.isscalar(idxlab):
      self.pdes[idxlab]=pde
    else:
      for idx in idxlab:
        self.pdes[idx]=pde
    
  
  def Assembly(self,**kwargs):
    dtype=kwargs.get('dtype',float)
    local=kwargs.get('local',False)
    Num=kwargs.get('Num',1)
    dom=kwargs.get('dom',True)
    physical=kwargs.get('physical',True)
    interface=kwargs.get('interface',False)
    Robin=kwargs.get('Robin',True)
    Dirichlet=kwargs.get('Dirichlet',True)
    A=sparse.csc_matrix((self.ndof,self.ndof),dtype=dtype)
    b=np.zeros((self.ndof,1))
    VFInd=getVFindices(Num,self.m,self.Th.nq)
    if dom:
      idxlab=self.Th.find(self.Th.d)
      A,b=Assembly_main(self,A,b,VFInd,idxlab,local)
    if Robin:
      A,b=Assembly_BC(self,A,b,VFInd,physical,interface,Assembly_Robin)
    if Dirichlet: 
      A,b=Assembly_BC(self,A,b,VFInd,physical,interface,Assembly_Dirichlet)
    
    return A,b
  
  def splitsol(self,X,**kwargs):
    Num=kwargs.get('Num',1)
    if self.m==1:
      return X
    VFInd=getVFindices(Num,self.m,self.Th.nq)
    x=self.m*[None]
    I=np.arange(self.Th.nq)
    for i in range(self.m):
      x[i]=X[VFInd(I,i)]
    return x
    
  def solve( self, **kwargs):
    # perm=lambda A: scipy.sparse.csgraph.reverse_cuthill_mckee(A)
    from fc_vfemp1.sparse import csr_one_rows,csr_sum_onerow
    from fc_vfemp1 import FEM,operators
    split=kwargs.pop('split',False)
    solver=kwargs.pop('solver', lambda A,b: splinalg.spsolve(A,b))
    perm=kwargs.pop('perm', None)
    condition=kwargs.get('condition',self.m*[None])
    A,b=self.Assembly(**kwargs)
    A,b=space_condition_pre(self,A,b,condition)
    #VFInd=getVFindices(1,self.m,self.Th.nq)
    #for i in range(self.m):
      #if condition[i]:
        #from fc_vfemp1.sparse import csr_one_rows,csr_sum_onerow
        
    #A.eliminate_zeros()
    b=b.reshape((b.shape[0],))
    if perm is not None:
      p=perm(A)
      x=np.ndarray((self.ndof,))
      x[p]=solver(A[p,::][::,p],b[p])
    else:
      x=solver(A,b)
    #Mass=None
    #I=np.arange(self.Th.nq)
    #Vone=np.ones((self.Th.nq,))
    #for i in range(self.m):
      #if condition[i]:
        #if Mass is None:
          #Mass=FEM.AssemblyP1(self.Th,operators.Lmass(self.Th.dim))
        #xI=x[VFInd(I,i)]
        #C=(1-np.dot(Mass*xI,Vone))/np.dot(Mass*Vone,Vone)
        #x[VFInd(I,i)]=xI+C
        #print(np.dot(Mass*x[VFInd(I,i)],Vone))
    x=space_condition_post(self,x,condition)
    
    if isinstance(x,tuple):
      info=x[1]
      x=x[0]
    #res=np.max(np.abs(A*x-b))
    #print('Residu : %.16f',res)
    if split:
      x=self.splitsol(x)
    return x
  
  
# A(tgr,tgc) <- A(tgr,tgc)+Ai
def sum_sub_mat(A,Ai,tgr,tgc):
  if Ai is None:
    return A
  if A.shape==Ai.shape:
    return A+Ai
  (I,J,K)=sparse.find(Ai)
  A+=sparse.csc_matrix((K,(tgr[I],tgc[J])),shape=A.shape)
  return A

def csr_zero_rows(A, rows_to_zero):
  if not isinstance(A, sparse.csc_matrix):
    raise ValueError('Matrix given must be of CSR format.')
  mask=np.in1d(A.indices, rows_to_zero)
  A.data[mask]=0
  #A.eliminate_zeros()
  #A.eliminate_zeros()
  
  
# A(tgr,:) <- 0 
# A(tgr,tgc) <- Ai
def dirichlet_mat(A,Ai,tgr,tgc):
  (I,J,K)=sparse.find(Ai)
  if tgr.ndim==0: # tgr is np.array(0) for example
    tgr=tgr.reshape((1,))
  if tgc.ndim==0:
    tgc=tgc.reshape((1,))
  csr_zero_rows(A,tgr)
  A+=sparse.csc_matrix((K,(tgr[I],tgc[J])),shape=A.shape)
  return A
  
def Assembly_main(self,A,b,VFInd,idxlab,local):
  block=self.m>1
  if np.isscalar(idxlab):
    idxlab=[idxlab]
  for i in idxlab:
    if self.pdes[i] is not None:
      Ai,bi=self.pdes[i].Assembly(self.Th.sTh[i],local=local,block=block)
      #print(Ai[0][1])
      if Ai is not None:
        if block:
          for l in np.arange(self.m):
            tgl=VFInd(self.Th.sTh[i].toParent,l)
            for s in np.arange(self.m):
              if Ai[l][s] is not None:
                tgs=VFInd(self.Th.sTh[i].toParent,s)
                A=sum_sub_mat(A,Ai[l][s],tgl,tgs)
            if bi[l] is not None:
              b[tgl]=b[tgl]+bi[l]
        else:
          tg=VFInd(self.Th.sTh[i].toParent,0)
          #print(tg)
          #A[tg,:][:,tg]=A[tg,:][:,tg]+Ai
          A=sum_sub_mat(A,Ai,tg,tg)
        
          if bi is not None:
            if bi.shape==b.shape:
              b+=bi
            else:
              b[tg]+=bi
  #print(A)
  return A,b

def Assembly_BC(self,A,b,VFInd,physical,interface,Assemblyfun):
  for dd in np.arange(self.Th.d-1,-1,-1):
    Allidxlab=self.Th.find(dd)
    if Allidxlab is not None:
      idxlab=np.array([],dtype=int)
      if physical:
        I=np.where((self.Th.sThlab[Allidxlab]>=0) & (self.Th.sThlab[Allidxlab]<10000))[0]
        idxlab=np.concatenate((idxlab,Allidxlab[I]))
      if interface:
        I=np.where((self.Th.sThlab(Allidxlab)<0) | (self.Th.sThlab(Allidxlab)>=10000))[0]
        idxlab=np.concatenate((idxlab,Allidxlab[I]))
      
      if len(idxlab)>0:
        [A,b]=Assemblyfun(self,A,b,VFInd,idxlab)
  return A,b

def Assembly_Robin(self,A,b,VFInd,idxlab):
  block=self.m>1
  #print(idxlab)
  for i in idxlab:
    #print(i)
    if self.pdes[i] is not None:
      #print(self.pdes[i])
      #print(self.Th.sTh[i])
      #Ai,bi=self.pdes[i].Assembly(self.Th.sTh[i],block=block)
      BB=self.pdes[i].Assembly(self.Th.sTh[i],block=block,local=True)
      if BB is not None:
        Ai,bi=BB
        if self.m>1:
          for l in np.arange(self.m):
            if self.pdes[i].delta[l]==1: # Robin
              tgl=VFInd(self.Th.sTh[i].toParent,l)   
              for k in np.arange(self.m):
                if Ai[l][k] is not None:
                  tgk=VFInd(self.Th.sTh[i].toParent,k)
                  #A[tgl,:][:,tgk]=A[tgl,:][:,tgk]+Ai[l,k]
                  A=sum_sub_mat(A,Ai[l][k],tgl,tgk)
              if bi[l] is not None:
                b[tgl]=b[tgl]+bi[l]
        else:
          if self.pdes[i].delta==1:
            tg=self.Th.sTh[i].toParent
            A=sum_sub_mat(A,Ai,tg,tg)
            if bi is not None:
              b[tg]=b[tg]+bi
  return A,b

def Assembly_Dirichlet(self,A,b,VFInd,idxlab):
  #print('Dirichlet')
  block=self.m>1
  for i in idxlab:
    #print(i)
    if self.pdes[i] is not None:
      BB=self.pdes[i].Assembly(self.Th.sTh[i],block=block,local=True)
      if BB is not None:
        Ai,bi=BB
        if self.m>1:
          for l in np.arange(self.m):
            if self.pdes[i].delta[l]==0: # Dirichlet
              tgl=VFInd(self.Th.sTh[i].toParent,l)   
              for k in np.arange(self.m):
                if Ai[l][k] is not None:
                  tgk=VFInd(self.Th.sTh[i].toParent,k)
                  A=dirichlet_mat(A,Ai[l][k],tgl,tgk)
              if bi[l] is not None:
                b[tgl]=bi[l]
        else:
          if self.pdes[i].delta==0:
            tg=self.Th.sTh[i].toParent
            A=dirichlet_mat(A,Ai,tg,tg)
            if bi is not None:
              b[tg]=bi
  return A,b

def Assembly_Dirichlet_pt(self,**kwargs):
  #print('Dirichlet')
  Num=kwargs.pop('Num',1)
  ndof=self.m*self.Th.nq
  gD=np.zeros((ndof,))
  VFInd=getVFindices(Num,self.m,self.Th.nq)
  idxlab=self.Th.find(self.Th.d-1)
  ID=np.array([],dtype=int)
  for i in idxlab:
    if self.pdes[i] is not None:
      tg=self.Th.sTh[i].toParent
      if self.m>1:
        for l in np.arange(self.m):
          if self.pdes[i].delta[l]==0: # Dirichlet
            tgl=VFInd(tg,l)
            gD[tgl]=self.Th.sTh[i].eval(self.pdes[i].f[l])
            ID=np.concatenate((ID,tgl))
      else:
        gD[tg]=self.Th.sTh[i].eval(self.pdes[i].f)
        ID=np.concatenate((ID,tg))
  ID=np.unique(ID)
  IDc=np.setdiff1d(np.arange(ndof),ID)      
  return ID,IDc,gD

def space_condition_pre(bvp,A,b,condition):
  from fc_vfemp1.sparse import csr_one_rows
  VFInd=getVFindices(1,bvp.m,bvp.Th.nq)
  for i in range(bvp.m):
    if condition[i] is not None:
      row=VFInd(0,i)
      csr_one_rows(A,row) 
      b[row]=1
  return A,b

def space_condition_post(bvp,x,condition):
  from fc_vfemp1 import FEM,operators
  Mass=None
  for i in range(bvp.m):
    if condition[i] is not None: # contains the value of \int_\Omega u_i dx  
      if Mass is None:
        Mass=FEM.AssemblyP1(bvp.Th,operators.Lmass(bvp.Th.dim))
        VFInd=getVFindices(1,bvp.m,bvp.Th.nq)
        I=np.arange(bvp.Th.nq)
        Vone=np.ones((bvp.Th.nq,))
      xI=x[VFInd(I,i)]
      C=(condition[i]-np.dot(Mass*xI,Vone))/np.dot(Mass*Vone,Vone)
      x[VFInd(I,i)]=xI+C
  return x

from fc_simesh.siMesh import siMesh
from fc_simesh.siMeshElt import siMeshElt
#from fc_vfemp1 import PDEelt as Pt
from fc_vfemp1.FEMtools import getVFindices
from fc_vfemp1.operators import Loperator,Hoperator,check_data,is_function_data,is_Loperator,is_Hoperator



import numpy as np
from fc_vfemp1.operators import StiffElasHoperator
from fc_vfemp1.BVP import BVP,PDE
from fc_simesh.siMesh import HyperCube
from fc_tools.others import isModuleFound

E=21e5;nu=0.45;L=20;N=10
mu= E/(2*(1+nu))
lam = E*nu/((1+nu)*(1-2*nu))
gam=lam+2*mu
L=5;N=7
print('------------------------------------------------------')
print('      3D elasticity BVP on [0,%g]x[0,1]x[0,1]'%L);
print('         on transform hypercube mesh')
print('             -div(sigma(u)) = f ')
print(' and boundary conditions :')
print('   * Dirichlet on [1]')
print('       u_1 = u_2 = 0 on [1],')
print('   * Neumann on [2,3,4,5,6]')
print('       sigma(u).n = 0 on [2,3,4,5,6]')
print(' with f=(0,0,-1)^t, lambda=%g, mu=%g'%(lam,mu))
print('------------------------------------------------------')
mapping=lambda q: np.array([L*q[0],q[1],q[2]])
Th=HyperCube(3,[L*N,N,N],mapping=mapping)
print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
print('2.a Setting 3D Elasticity B.V.P. problem')

Hop=StiffElasHoperator(3,lam,mu)
pde=PDE(Op=Hop,f=[0,0,-1])
bvp=BVP(Th,pde=pde)
bvp.setDirichlet(1,[0,0,0])

print('2.b Solving 3D elasticity BVP')
U=bvp.solve(split=True)

Th=bvp.Th

normU=np.sqrt(U[0]**2+U[1]**2+U[2]**2)

if isModuleFound('fc_simesh_mayavi'):
  from mayavi import mlab
  import fc_simesh_mayavi.siMesh as simlab
  scale=2000
  if len(mlab.get_engine().scenes)>0:
    mlab.close(all=True)

  mlab.figure(1)
  simlab.plotmesh(Th,d=2,legend=True)
  
  mlab.figure(2)
  simlab.plotmesh(Th,color='blue',opacity=0.01)
  simlab.plotmesh(Th,color='red',move=[scale*U[0],scale*U[1],scale*U[2]])
  
  mlab.figure(3)
  simlab.plot(Th,normU,d=2)
  simlab.plotiso(Th,normU,d=2,contours=10,color='White',line_width=1)
  simlab.plotmesh(Th,d=1,color='black')
  mlab.colorbar(title='|U|')

if isModuleFound('fc_simesh_matplotlib'):  
  import matplotlib.pyplot as plt
  import fc_simesh_matplotlib.siMesh as siplt
  from fc_tools.matplotlib import set_axes_equal,DisplayFigures
  plt.close('all')
  plt.ion()
  DisplayFigures(nfig=3)
  plt.figure(1)
  siplt.plotmesh(Th,d=2,legend=True)
  set_axes_equal()
  
  plt.figure(2)
  siplt.plotmesh(Th,color='blue',alpha=0.05)
  siplt.plotmesh(Th,color='red',move=[scale*U[0],scale*U[1],scale*U[2]],alpha=0.1)
  plt.axis('off');set_axes_equal()
  
  plt.figure(3)
  hp=siplt.plot(Th,normU,d=2)
  plt.colorbar(hp,label='|u|')
  siplt.plotmesh(Th,d=1,color='black')
  plt.axis('off');set_axes_equal()
  


# Run with: from fc_vfemp1.examples import BVPStationaryConvectionDiffusion3D01
import numpy as np
from fc_vfemp1.examples.set_examples import setBVPStationaryConvectionDiffusion3D01
bvp,info=setBVPStationaryConvectionDiffusion3D01(20,verbose=2)
print('3. Solving BVP problem')
u=bvp.solve()

print('4. Plotting mesh and solution')
from fc_tools.others import isModuleFound
Th=bvp.Th
if isModuleFound('fc_simesh_mayavi'):
  from mayavi import mlab
  import fc_simesh_mayavi.siMesh as simlab
  if len(mlab.get_engine().scenes)>0:
    mlab.close(all=True)
  mlab.figure(1)
  simlab.plotmesh(Th,legend=True)
  mlab.figure(2)
  simlab.plotmesh(Th,d=2,legend=True)
  mlab.figure(3)
  simlab.plotmesh(Th,d=2,legend=True,labels=[10,20,21,100,101])
  mlab.figure(4)
  simlab.plot(Th,u,line_width=1)
  mlab.colorbar(title='u',orientation='vertical')
  mlab.figure(5)
  simlab.plot(Th,u,d=2,labels=[10,20,21,100,101])
  simlab.plotiso(Th,u,d=2,labels=[10,20,21,100,101],contours=10,color='White')
  mlab.colorbar(title='u',orientation='vertical')
  mlab.figure(6)
  simlab.slice(Th,u,origin=(0,0,1),normal=(0,1,1))
  simlab.sliceiso(Th,u,origin=(0,0,1),normal=(0,1,1),contours=10,color='White')
  simlab.plot(Th,u,d=2,labels=[100,101])
  simlab.plotiso(Th,u,d=2,labels=[100,101],contours=10,color='White')
  simlab.plot(Th,u,d=2,labels=[10,20,21],representation='wireframe',line_width=1)
  mlab.colorbar(title='u',orientation='vertical')
  mlab.view(-49.8,73.4,7)
  
if isModuleFound('fc_simesh_matplotlib'):
  import matplotlib.pyplot as plt
  import fc_simesh_matplotlib.siMesh as siplt
  from fc_tools.matplotlib import set_axes_equal,DisplayFigures
  plt.close('all')
  plt.ion()
  DisplayFigures(nfig=5)
  plt.figure(1)
  siplt.plotmesh(Th,legend=True)
  set_axes_equal()
  plt.figure(2)
  siplt.plotmesh(Th,d=2,legend=True)
  set_axes_equal()
  plt.figure(3)
  siplt.plotmesh(Th,d=2,labels=[10,20,21,100,101],legend=True)
  set_axes_equal()
  plt.figure(4)
  hp=siplt.plot(Th,u)
  plt.colorbar(hp,label='u')
  plt.axis('off');set_axes_equal()
  plt.figure(5)
  hp=siplt.plot(Th,u,d=2,labels=[10,20,21,100,101])
  plt.colorbar(hp,label='u')
  plt.axis('off');set_axes_equal()
  



import fc_oogmsh
from fc_vfemp1.operators import Loperator,Hoperator
from fc_vfemp1.BVP import BVP,PDE
from fc_vfemp1.sys import get_geo
from fc_simesh.siMesh import siMesh,HyperCube
from fc_vfemp1.common import PrintCopyright 

import inspect,types,os

import numpy as np
from numpy import cos,sin,exp,pi
import time,socket
import matplotlib.pyplot as plt
from math import * 

def func2str(u):
  if isinstance(u,types.LambdaType):
    Str=inspect.getsource(u).replace('\n','')
    i=Str.find('=')+1
    return Str[i::]
    #return Str
  else:
    return str(u)

def setBVPStationaryConvectionDiffusion2D01(N,**kwargs):
  env=fc_oogmsh.sys.environment()
  verbose=kwargs.get('verbose',0)
  if (N==0) and (verbose<10):
    verbose=10  
  geodir=kwargs.get('geodir',env.geo_dir)
  meshdir=kwargs.get('meshdir',env.mesh_dir)
  geofile='disk3holes'
  af=lambda x,y: 0.1+(y-0.5)*(y-0.5)
  Vx=lambda x,y: -10*y
  Vy=lambda x,y: 10*x
  b=0.01;g2=4;g4=-4;
  f=lambda x,y: -200.0*exp(-((x-0.75)**2+y**2)/(0.1))
  
  if verbose>1:
    PrintCopyright()
    print('------------------------------------------------------') 
    print('      2D stationary convection-diffusion BVP')
    print('         on GMSH mesh from %s.geo\n'%geofile)
    print('               - div(a grad u) + <V,grad u> + b u = f')
    print(' and boundary conditions :')
    print('   * Dirichlet on [2,4,20,21]')
    print('       u = 4 on [2], -4 on [4] and 0 on [20,21]  ')
    print('   * Neumann on [1,3,10]')
    print('       du/dn = 0 on [2,3,10]')
    print(' with ')
    print('   a = %s'%func2str(af))
    print('   V = [%s,%s] '%(func2str(Vx),func2str(Vy)))
    print('   b = %g'% b)
    print('   f = %s'%func2str(f))
    print('------------------------------------------------------\n')
    if verbose>=10:
      return
  if verbose>=2:
    print('*** Building the mesh using GMSH')
  
  meshfile=fc_oogmsh.gmsh.buildmesh2d(geofile,N,force=True,verbose=0)
  Th=siMesh(meshfile)
  if verbose>=2:
    print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
    print('*** Setting 2D Condenser BVP problem')
 
  pde=PDE(Op=Loperator(dim=2,d=2,A=[[af,None],[None,af]],c=[Vx,Vy],a0=b),f=f)
  bvp=BVP(Th,pde=pde)
  bvp.setDirichlet(2,g2)
  bvp.setDirichlet(4,g4)
  bvp.setDirichlet(20,0)
  bvp.setDirichlet(21,0)
  info=dict(solex=None,geofile=geofile)
  return bvp,info

def setBVPStationaryConvectionDiffusion3D01(N,**kwargs):
  verbose=kwargs.get('verbose',0)
  if (N==0) and (verbose<10):
    verbose=10  
  geofile='cylinder3holes'
  af=lambda x,y,z: 0.7+z/10
  Vx=lambda x,y,z: -10*y;Vy=lambda x,y,z: 10*x;Vz=lambda x,y,z: 10*z
  f=lambda x,y,z: -800.0*np.exp(-10*((x-0.65)**2+y*y+(z-0.5)**2))+\
                   800.0*np.exp(-10*((x+0.65)**2+y*y+(z-0.5)**2))
  b=0.01              
  
  (geodir,geofile)=get_geo(3,3,geofile)
  env=fc_oogmsh.sys.environment()
  geodir=kwargs.get('geodir',geodir)#env.geo_dir)
  meshdir=kwargs.get('meshdir',env.mesh_dir)
  
  if verbose>1:
    PrintCopyright()
    print('------------------------------------------------------') 
    print('      3D stationary convection-diffusion BVP')
    print('         on GMSH mesh from %s.geo\n'%geofile)
    print('               - div(a grad u) + <V,grad u> + b u = f')
    print(' and boundary conditions :')
    print('   * Robin on [20,21]')
    print('       du/dn + u= +0.05 on [20]')
    print('       du/dn + u= -0.05 on [21]')
    print('   * Neumann otherwise')
    print('       du/dn = 0')
    print(' with ')
    print('   a = %s'%func2str(af))
    print('   V = [%s,%s,%s] '%(func2str(Vx),func2str(Vy),func2str(Vz)))
    print('   b = %g'% b)
    print('   f = %s'%func2str(f))
    print('------------------------------------------------------\n')
    if verbose>=10:
      return
  if verbose>=2:
    print('1. Building the mesh using GMSH')
  import os
  meshfile=fc_oogmsh.gmsh.buildmesh3d(geodir+os.sep+geofile+'.geo',N, force=True,verbose=0,meshdir=meshdir)
  Th=siMesh(meshfile)
  if verbose>=2:
    print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
    print('2. Setting 3D Stationary convection/diffusion BVP problem')
 
  pde=PDE(Op=Loperator(dim=3,d=3,A=[[af,None,None],[None,af,None],[None,None,af]], c=[Vx,Vy,Vy],a0=b),f=f)
  bvp=BVP(Th,pde=pde)
  bvp.setRobin(20,+0.05,ar=1)
  bvp.setRobin(21,-0.05,ar=1)
  info=dict(solex=None,geofile=geofile)
  return bvp,info

def setBVPPoisson2D_ex01(N,**kwargs):
  env=fc_oogmsh.sys.environment()
  verbose=kwargs.get('verbose',0)
  if (N==0) and (verbose<10):
    verbose=10  
  geodir=kwargs.get('geodir',env.geo_dir)
  meshdir=kwargs.get('meshdir',env.mesh_dir)
  uex=lambda x,y: cos(x-y)*sin(x+y)+exp(-(x**2+y**2))
  f=lambda x,y: -4*x**2*exp(-x**2-y**2) - 4*y**2*exp(-x**2-y**2) + 4*cos(x-y)*sin(x+y) + 4*exp(-x**2-y**2)
  
  if verbose>=1:
    PrintCopyright()
    print('------------------------------------------------------')
    print('      2D Poisson B.V.P. problem on unit square:')
    print('               -Laplacian(u) = f')
    print(' with  ')
    print('    * Dirichlet boundary conditions on [1,2,3,4]')
    print(' Exact solution :\n   %s'%func2str(uex))
    print('------------------------------------------------------')
    if verbose>=10:
      return

  if verbose>=2:
    print('1. Setting the mesh using HyperCube function')
  
  Th=HyperCube(2,N)
  if verbose>=2:
    print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
    print('2. Setting 2D Condenser BVP problem')
    
  Lop=Loperator(dim=2,d=2,A=[[1,None],[None,1]])
  pde=PDE(Op=Lop,f=f)
  bvp=BVP(Th,pde=pde)
  bvp.setDirichlet( 1, uex)
  bvp.setDirichlet( 2, uex)
  bvp.setDirichlet( 3, uex)
  bvp.setDirichlet( 4, uex)
  
  info=dict(solex=uex,f=f)
  return bvp,info

def setBVPPoisson2D_ex02(N,**kwargs):
  env=fc_oogmsh.sys.environment()
  verbose=kwargs.get('verbose',0)
  if (N==0) and (verbose<10):
    verbose=10  
  geodir=kwargs.get('geodir',env.geo_dir)
  meshdir=kwargs.get('meshdir',env.mesh_dir)
  uex=lambda x,y: cos(4*pi*(x**2 + y**2))*exp(-2*(x**2 +y**2))
  f=lambda x,y: ( 64*pi**2*x**2*cos(4*pi*(x**2 + y**2))*exp(-2*x**2 - 2*y**2) +
                  64*pi**2*y**2*cos(4*pi*(x**2 + y**2))*exp(-2*x**2 - 2*y**2) - 
                  64*pi*x**2*exp(-2*x**2 - 2*y**2)*sin(4*pi*(x**2 + y**2)) -
                  64*pi*y**2*exp(-2*x**2 - 2*y**2)*sin(4*pi*(x**2 + y**2)) - 
                  16*x**2*cos(4*pi*(x**2 + y**2))*exp(-2*x**2 - 2*y**2) - 
                  16*y**2*cos(4*pi*(x**2 + y**2))*exp(-2*x**2 - 2*y**2) +
                  16*pi*exp(-2*x**2 - 2*y**2)*sin(4*pi*(x**2 + y**2)) +
                  8*cos(4*pi*(x**2 + y**2))*exp(-2*x**2 - 2*y**2)
                )
  if verbose>=1:
    PrintCopyright()
    print('------------------------------------------------------')
    print('      2D Poisson B.V.P. problem on unit square:')
    print('               -Laplacien(u) = f')
    print(' with  ')
    print('    * Dirichlet boundary conditions on [1,2,3,4]')
    print(' Exact solution :\n   %s'%func2str(uex))
    print('------------------------------------------------------')
    if verbose>=10:
      return

  if verbose>=2:
    print('*** Setting the mesh using HyperCube function')
  
  Th=HyperCube(2,N)
  if verbose>=2:
    print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
    print('*** Setting 2D Condenser BVP problem')
    
  Lop=Loperator(dim=2,d=2,A=[[1,None],[None,1]])
  pde=PDE(Op=Lop,f=f)
  bvp=BVP(Th,pde=pde)
  bvp.setDirichlet( 1, uex)
  bvp.setDirichlet( 2, uex)
  bvp.setDirichlet( 3, uex)
  bvp.setDirichlet( 4, uex)
  
  info=dict(solex=uex,f=f)
  return bvp,info

def setBVPPoisson2D_ex03(N,**kwargs):
  env=fc_oogmsh.sys.environment()
  verbose=kwargs.get('verbose',0)
  if (N==0) and (verbose<10):
    verbose=10  
  geodir=kwargs.get('geodir',env.geo_dir)
  meshdir=kwargs.get('meshdir',env.mesh_dir)
  uex=lambda x,y: cos(2*x+y)
  f=lambda x,y: 5*cos(2*x+y)
  gradu=[lambda x,y: -2*sin(2*x+y),lambda x,y: -sin(2*x+y)]
  ar3 = lambda x,y: 1+x**2+y**2
  if verbose>=1:
    PrintCopyright()
    print('------------------------------------------------------')
    print('      2D Poisson B.V.P. problem on unit square:')
    print('               -Laplacien(u) = f')
    print(' with  ')
    print(' Exact solution : %s'%func2str(uex))
    print(' and boundary conditions :')
    print('   * Dirichlet on [1,2]')
    print('       u = gD')
    print('   * Robin boundary condition on [3]')
    print('       du/dn + ar*u = gR')
    print('         with ar=%s'%func2str(ar3))
    print('   * Neumann boundary condition on [4]')
    print('       du/dn = gN')
    print('------------------------------------------------------');
    if verbose>=10:
      return

  if verbose>=2:
    print('*** Setting the mesh using HyperCube function')
  
  Th=HyperCube(2,N)
  if verbose>=2:
    print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
    print('*** Setting 2D Condenser BVP problem')
    
  Lop=Loperator(dim=2,d=2,A=[[1,None],[None,1]])
  pde=PDE(Op=Lop,f=f)
  bvp=BVP(Th,pde=pde)
  bvp.setDirichlet( 1, uex)
  bvp.setDirichlet( 2, uex)
  bvp.setRobin( 3,lambda x,y: -gradu[1](x,y)+ar3(x,y)*uex(x,y) , ar=ar3)
  bvp.setRobin( 4, gradu[1])
  
  info=dict(solex=uex,f=f)
  return bvp,info

def setBVPPoisson3D_ex01(N,**kwargs):
  env=fc_oogmsh.sys.environment()
  verbose=kwargs.get('verbose',0)
  if (N==0) and (verbose<10):
    verbose=10  
  geodir=kwargs.get('geodir',env.geo_dir)
  meshdir=kwargs.get('meshdir',env.mesh_dir)
  uex=lambda x,y,z: cos(4*x-3*y+5*z)
  f=lambda x,y,z: 50*uex(x,y,z)
  gradu=[lambda x,y,z: -4*sin(4*x-3*y+5*z), lambda x,y,z: 3*sin(4*x-3*y+5*z), lambda x,y,z: -5*sin(4*x-3*y+5*z)]
  ar = 1
  if verbose>=1:
    PrintCopyright()
    print('------------------------------------------------------')
    print('      3D Poisson B.V.P. problem on unit square:')
    print('               -Laplacien(u) = f')
    print(' with  ')
    print(' Exact solution : %s'%func2str(uex))
    print(' and boundary conditions :')
    print('   * Dirichlet on [1,3,5]')
    print('       u = gD')
    print('   * Robin boundary condition on [2,4]')
    print('       du/dn + ar*u = gR')
    print('         with ar=%s'%func2str(ar))
    print('   * Neumann boundary condition on [6]')
    print('       du/dn = gN')
    print('------------------------------------------------------');
    if verbose>=10:
      return

  if verbose>=2:
    print('1. Setting the mesh using HyperCube function')
  
  Th=HyperCube(3,N)
  if verbose>=2:
    print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
    print('2. Setting 3D Poisson B.V.P. problem')
    
  Lop=Loperator(dim=3,d=3,A=[[1,None,None],[None,1,None],[None,None,1]])
  pde=PDE(Op=Lop,f=f)
  bvp=BVP(Th,pde=pde)
  for lab in [1,3,5]:
    bvp.setDirichlet( lab, uex)
  bvp.setRobin( 2,lambda x,y,z: gradu[0](x,y,z)+ar*uex(x,y,z) , ar=ar)
  bvp.setRobin( 4,lambda x,y,z: gradu[1](x,y,z)+ar*uex(x,y,z) , ar=ar)
  bvp.setRobin( 6,gradu[2])
  
  info=dict(solex=uex,f=f)
  return bvp,info


def setBVPElasticity2d_ex01(N,**kwargs):
  verbose=kwargs.get('verbose',0)
  if (N==0) and (verbose<10):
    verbose=10  
  E=kwargs.get('E',21e5)
  nu=kwargs.get('nu',0.45)
  mu= E/(2*(1+nu))
  lam = E*nu/((1+nu)*(1-2*nu))
  gam=lam+2*mu
  uex=[lambda x,y: cos(2*x+y), lambda x,y: sin(x-3*y)]
  f=[lambda x,y: 4*lam*cos(2*x + y) + 9*mu*cos(2*x + y) + 3*lam*sin(-x + 3*y) + 3*mu*sin(-x + 3*y) ,
     lambda x,y: 2*lam*cos(2*x + y) + 2*mu*cos(2*x + y) - 9*lam*sin(-x + 3*y) - 19*mu*sin(-x + 3*y)]
  g2=[lambda x,y: -3*lam*cos(-x + 3*y) - 2*lam*sin(2*x + y) - 4*mu*sin(2*x + y), 
      lambda x,y: mu*(cos(-x + 3*y) - sin(2*x + y))]
  if verbose>=1:
    PrintCopyright()
    print('------------------------------------------------------') 
    print('      2D Elasticity B.V.P. problem on unit square:')
    print('         - div sigma(u) = f')
    print(' with  * Dirichlet boundary conditions on [1,3,4]')
    print('       * Neumann boundary conditions on [2]')
    print(' and Young modulus E=%g, Poisson ration nu=%g\n'%(E,nu)) 
    print(' exact sol. : [%s,%s]'%(func2str(uex[0]),func2str(uex[1])))
    print('------------------------------------------------------')
    if verbose==10: 
      return
  
  if verbose>=2:
    print('1. Setting the mesh using HyperCube function')
  
  Th=HyperCube(2,N)
  if verbose>=2:
    print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
    print('2. Setting 2D Elasticity B.V.P. problem')
  Hop=Hoperator(dim=2,d=2,m=2)
  Hop.H[0][0]=Loperator(d=2,A=[[gam,None],[None,mu]]) 
  Hop.H[0][1]=Loperator(d=2,A=[[None,lam],[mu,None]]) 
  Hop.H[1][0]=Loperator(d=2,A=[[None,mu],[lam,None]]) 
  Hop.H[1][1]=Loperator(d=2,A=[[mu,None],[None,gam]])
  
  pde=PDE(Op=Hop,f=f)
  bvp=BVP(Th,pde=pde)
  
  for lab in [1,3,4]:
    bvp.setDirichlet(lab,uex)
  bvp.setRobin(2,g2)
  
  info=dict(solex=uex,geofile='hypercube')
  return bvp,info

def setBVPElectrostatic2D01(N,**kwargs):
  verbose=kwargs.get('verbose',0)
  if (N==0) and (verbose<10):
    verbose=10  
  sigma1=kwargs.get('sigma1',1)
  sigma2=kwargs.get('sigma2',10)
  (geodir,geofile)=get_geo(2,2,'square4holes6dom')
  env=fc_oogmsh.sys.environment()
  geodir=kwargs.get('geodir',geodir)
  meshdir=kwargs.get('meshdir',env.mesh_dir)
  
  if verbose>1:
    PrintCopyright()
    print('------------------------------------------------------') 
    print('      2D electrostatic BVP problem ')
    print('        on square [-1,1]x[-1,1] with ')
    print('        4 holes and 6 domains ')
    print('------------------------------------------------------') 
    print(' PDE :')
    print('    * -div(sigma grad(phi)) = 0   on domain [2,4,6,8,10,20]')
    print(' Local electrical conductivity : ')
    print('    * sigma=sigma1=%g on domain [10]'%sigma1)
    print('    * sigma=sigma2=%g on domain [2,4,6,8,20]'%sigma2)
    print(' Dirichlet boundary condition on boundaries [1,3,5,7]')
    print('    * on [1,5] : phi = 12')
    print('    * on [3,7] : phi = 0')
    print(' Neumann boundary condition on boundary [20]')
    print('    * on [20] : sigma d(phi)/dn = 0')
    print('------------------------------------------------------') 
    if verbose>=10:
      return
  if verbose>=2:
    print('1. Building the mesh using GMSH')
  
  meshfile=fc_oogmsh.gmsh.buildmesh2d(geodir+os.sep+geofile+'.geo',N,force=True)
  Th=siMesh(meshfile)
  if verbose>=2:
    print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
    print('2. Setting 2D Electrostatic BVP')
 
  pde=PDE(Op=Loperator(dim=2,A=[[sigma2,None],[None,sigma2]]))
  bvp=BVP(Th,pde=pde)
  Lop=Loperator(dim=2,A=[[sigma1,None],[None,sigma1]])
  bvp.setPDE(PDE(Op=Lop),labels=10)              
  bvp.setDirichlet(1,12)
  bvp.setDirichlet(3,0)
  bvp.setDirichlet(5,12)
  bvp.setDirichlet(7,0)
  info=dict(solex=None,geofile=geofile,sigma1=sigma1,sigma2=sigma2)
  return bvp,info

def setBVPCondenser2D01(N,**kwargs):
  verbose=kwargs.get('verbose',0)
  if (N==0) and (verbose<10):
    verbose=10  
  u1=kwargs.get('u1',1)
  u2=kwargs.get('u2',-1)
  geofile='condenser'
  (geodir,geofile)=get_geo(2,2,geofile)
  env=fc_oogmsh.sys.environment()
  geodir=kwargs.get('geodir',geodir)
  meshdir=kwargs.get('meshdir',env.mesh_dir)
  
  if verbose>1:
    PrintCopyright()
    print('------------------------------------------------------') 
    print('      2D condenser BVP problem ')
    print('        on mesh build from gmsh and %s file '%geofile)
    print('------------------------------------------------------') 
    print(' PDE :')
    print('    * -Laplacian(u) = 0   on domain [1]')
    print(' Dirichlet boundary condition on boundaries [1,98,99]')
    print('    * on [0]  : u = 0')
    print('    * on [98] : u = %d'%u1)
    print('    * on [99] : u = %d'%u2)
    print('------------------------------------------------------') 
    if verbose>=10:
      return
  if verbose>=2:
    print('*** Building the mesh using GMSH')
  
  meshfile=fc_oogmsh.gmsh.buildmesh2d(geodir+os.sep+geofile+'.geo',N,force=True)
  Th=siMesh(meshfile)
  if verbose>=2:
    print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
    print('*** Setting 2D Condenser BVP')
 
  pde=PDE(Op=Loperator(dim=2,A=[[1,None],[None,1]]))
  bvp=BVP(Th,pde=pde)
  bvp.setDirichlet(1,0)
  bvp.setDirichlet(98,u1)
  bvp.setDirichlet(99,u2)
  info=dict(solex=None,geofile=geofile,u1=u1,u2=u2)
  return bvp,info


import numpy as np

from fc_vfemp1.operators import Loperator
from fc_vfemp1.examples.set_examples import setBVPElectrostatic2D01
from fc_vfemp1 import FEM
from fc_tools.others import isModuleFound
geodir='/home/cuvelier/Travail/Recherch/python/fc-vfemp1/geodir/2d'

print(globals())
if 'sigma1' in globals():
  print('sigma1=%g'%sigma1)
  
print(locals())
if 'sigma1' in locals():
  print('sigma1=%g'%sigma1)

bvp,info=setBVPElectrostatic2D01(80,verbose=2,geodir=geodir,sigma1=10)
u=bvp.solve()

Th=bvp.Th
E=np.zeros((2,Th.nq))
Lop=Loperator(dim=2,d=2,c=[1,0])
E[0]=FEM.Apply(Lop,Th,u)#,solver=solver,perm=perm)
Lop=Loperator(dim=2,d=2,c=[0,1])
E[1]=FEM.Apply(Lop,Th,u)#,solver=solver,perm=perm)

ENorm=np.sqrt(E[0]**2+E[1]**2)

if isModuleFound('fc_simesh_mayavi'):
  from mayavi import mlab
  import fc_simesh_mayavi.siMesh as simlab
  if len(mlab.get_engine().scenes)>0:
    mlab.close(all=True)

  mlab.figure(1)
  simlab.plotmesh(Th,legend=True)
  mlab.view(0,0)

  mlab.figure(2)
  simlab.plotmesh(Th,d=1,legend=True,line_width=2)
  simlab.plotmesh(Th,color='LightGray',opacity=0.3)
  mlab.view(0,0)

  mlab.figure(3)
  simlab.plot(Th,u,colormap='viridis',plane=False)
  simlab.plotiso(Th,u,contours=25,color='white',line_width=2,plane=False)
  #mlab.view(0,0)
  cb=mlab.colorbar(title='u',orientation='vertical')
  cb.lut_mode='viridis'

  mlab.figure(4)
  simlab.plotiso(Th,u,contours=25,colormap='viridis')
  mlab.view(0,0)
  mlab.colorbar(title='u',orientation='vertical')
  simlab.plotmesh(Th,color='LightGray',opacity=0.05)
  simlab.plotmesh(Th,d=1,color='black',line_width=1.5)

  mlab.figure(5)
  simlab.plot(Th,ENorm,d=2,colormap='jet')
  simlab.plotiso(Th,ENorm,d=2,contours=20,color=(1,1,1),line_width=1)
  simlab.plotmesh(Th,d=1,color='black',line_width=1.5)
  mlab.view(0,0)

  mlab.figure(6)
  simlab.quiver(Th,E,scale_factor=0.3e-3)
  simlab.plotmesh(Th,d=1,color='black',line_width=1.5)
  mlab.view(0,0)

if isModuleFound('fc_simesh_matplotlib'):  
#if isMatplotlib():
  import matplotlib.pyplot as plt
  import fc_simesh_matplotlib.siMesh as siplt
  from fc_tools.matplotlib import set_axes_equal,DisplayFigures
  plt.close('all')
  plt.ion()
  DisplayFigures(nfig=6)
  plt.figure(1)
  siplt.plotmesh(Th,legend=True)
  set_axes_equal()
  
  plt.figure(2)
  siplt.plotmesh(Th,d=1,legend=True,linewidth=2)
  siplt.plotmesh(Th,color='LightGray',alpha=0.3)
  set_axes_equal()

  plt.figure(3)
  hp=siplt.plot(Th,u,colormap='viridis',plane=False)
  plt.colorbar(hp,label='u')
  #siplt.plotiso(Th,u,contours=25,color='white',linewidth=2,plane=False)
  plt.axis('off')#;set_axes_equal()

  plt.figure(4)
  siplt.plotiso(Th,u,contours=25,colormap='viridis')
  plt.colorbar(label='u')
  siplt.plotmesh(Th,color='LightGray',alpha=0.05)
  siplt.plotmesh(Th,d=1,color='black',linewidth=1.5)
  plt.axis('off');set_axes_equal()

  plt.figure(5)
  siplt.plot(Th,ENorm,d=2,colormap='jet')
  plt.colorbar(label='$\|E\|$ ')
  siplt.plotiso(Th,ENorm,d=2,contours=20,color=(1,1,1),linewidth=1)
  siplt.plotmesh(Th,d=1,color='black',linewidth=1.5)
  plt.axis('off');set_axes_equal()

  plt.figure(6)
  siplt.quiver(Th,E,scale_factor=0.3e-3)
  plt.colorbar(label='$E$ ')
  siplt.plotmesh(Th,d=1,color='black',linewidth=1.5)
  plt.axis('off');set_axes_equal()


import numpy as np

import fc_oogmsh
from fc_vfemp1.operators import Loperator,Hoperator
from fc_vfemp1.BVP import BVP,PDE
from fc_simesh.siMesh import siMesh
from fc_vfemp1.examples.set_examples import func2str
from fc_tools.others import isModuleFound
from fc_tools.mayavi import mlab_latex3D

__all__ = ['method1','method2']

def set_datas(N):
  dim=2
  af=lambda x,y: 0.1+y**2
  gD=lambda x,y: 20*y
  b=0.01
  geofile='disk5holes.geo'

  print('------------------------------------------------------') 
  print('      2D stationary heat and flow velocity BVPs')
  print('         on GMSH mesh from %s.geo'%geofile)
  print(' 1) 2D velocity potential BVP, (unknow phi)')
  print('               - Delta phi = 0')
  print('    and boundary conditions :')
  print('      * Dirichlet on [20,21]')
  print('          phi = -20 on [21], 20 on [20]')
  print('      * Neumann on [1,22,23]')
  print('          du/dn = 0 on [1,22,23]')
  print('    (Velocity V = grad phi ) ')
  print(' 2) 2D stationary heat with potential flow BVP, (unknow u)')
  print('               - div(a grad u) + <V,grad u> + b u = f')
  print(' and boundary conditions :')
  print('   * Dirichlet on [21,22,23]')
  print('       u = gD on [21], 0 on [22] and 0 on [23]  ')
  print('   * Neumann on [1,10,20]')
  print('       du/dn = 0 on [1,10,20]')
  print(' with ')
  print('   a = %s '%func2str(af))
  print('   b = %s '%func2str(b))
  print('   gD = %s '%func2str(gD))
  print('------------------------------------------------------')
  
  print('*** Building the mesh using GMSH')
  meshfile=fc_oogmsh.gmsh.buildmesh2d(geofile,N,force=True)
  Th=siMesh(meshfile)
  print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
  return dim,af,gD,b,geofile,Th
  
def method1(N=60):
  dim,af,gD,b,geofile,Th=set_datas(N)
  print('*** Setting 2D velocity potential BVP')
  Lop=Loperator(dim=dim,A=[[1,None],[None,1]])
  bvpVelocityPotential=BVP(Th,pde=PDE(Op=Lop))
  bvpVelocityPotential.setDirichlet(20,+20.)
  bvpVelocityPotential.setDirichlet(21,-20.)
  print('*** Solving 2D velocity potential BVP')
  phi=bvpVelocityPotential.solve()
  print('*** Setting 2D potential flow operator')
  Hop=Hoperator(dim=dim,m=2)
  Hop.H[0][0]=Loperator(dim=dim,c=[1,0])
  Hop.H[1][1]=Loperator(dim=dim,c=[0,1])
  print('*** Applying 2D potential flow operator')
  V=Hop.apply(Th,[phi,phi])
  print('*** Setting 2D stationary heat BVP with potential flow')
  Lop=Loperator(dim=dim,A=[[af,None],[None,af]],c=V,a0=b)
  bvpHeat=BVP(Th,pde=PDE(Op=Lop))
  bvpHeat.setDirichlet(21,gD)
  bvpHeat.setDirichlet(22, 0)
  bvpHeat.setDirichlet(23, 0)
  print('*** Solving 2D stationary heat BVP with potential flow')
  u=bvpHeat.solve()

  if isModuleFound('fc_simesh_mayavi'):
    print('*** Graphic representations using Mayavi')
    mayavi_plot(Th,phi,V,u)
  if isModuleFound('fc_simesh_matplotlib'):  
    print('*** Graphic representations using Matplotlib')
    matplotlib_plot(Th,phi,V,u)
  return Th,phi,V,u

def method2(N=60):
  dim,af,gD,b,geofile,Th=set_datas(N)
  print('*** Setting 2D potential velocity/flow BVP')
  Hop=Hoperator(dim=dim,m=3)
  Hop.H[0][1]=Loperator(dim=dim,b=[-1,0])
  Hop.H[0][2]=Loperator(dim=dim,b=[0,-1])
  Hop.H[1][0]=Loperator(dim=dim,c=[-1,0])
  Hop.H[1][1]=Loperator(dim=dim,a0=1)
  Hop.H[2][0]=Loperator(dim=dim,c=[0,-1])
  Hop.H[2][2]=Loperator(dim=dim,a0=1)
  bvpFlow=BVP(Th,pde=PDE(Op=Hop))
  bvpFlow.setDirichlet(20,20,comps=[0])
  bvpFlow.setDirichlet(21,-20,comps=[0])
  print('*** Solving 2D potential velocity/flow BVP')
  U=bvpFlow.solve(split=True) 
  print('*** Setting 2D stationary heat BVP with potential flow')
  V=np.array([U[1],U[2]])
  phi=U[0] 
  Lop=Loperator(dim=d,d=d,A=[[af,None],[None,af]],c=V,a0=b)
  bvpHeat=BVP(Th,PDEelt(Lop))
  bvpHeat.setDirichlet(21,gD)
  bvpHeat.setDirichlet(22, 0)
  bvpHeat.setDirichlet(23, 0)
  print('*** Solving 2D stationary heat BVP with potential flow')
  u=bvpHeat.solve()
  
  if isModuleFound('fc_simesh_mayavi'):
    print('*** Graphic representations using Mayavi')
    mayavi_plot(Th,phi,V,u)
  if isModuleFound('fc_simesh_matplotlib'):  
    print('*** Graphic representations using Matplotlib')
    matplotlib_plot(Th,phi,V,u)
  return Th,phi,V,u

def mayavi_plot(Th,phi,V,u):
  from mayavi import mlab
  import fc_simesh_mayavi.siMesh as simlab
  from fc_tools.mayavi import title
  v2=np.sqrt(V[0]**2+V[1]**2)
  mv2=np.max(v2)

  if len(mlab.get_engine().scenes)>0:
    mlab.close(all=True)
  mlab.figure(1)
  simlab.plotmesh(Th,color='LightGray',opacity=0.3)
  simlab.plotmesh(Th,d=1,legend=True)
  mlab.view(0,0)
  
  textproperty={'bold':True,'font_size':16,'shadow':True}
  latex_property={'color':'Black','center':(0,1.2,0)}

  mlab.figure(2)
  simlab.plot(Th,u,colormap='jet')
  simlab.plotiso(Th,u,contours=np.arange(1,17),color='white',line_width=1,opacity=0.5)
  mlab.view(0,0)
  mlab.colorbar(title='u',orientation='vertical')
  mlab_latex3D('2D stationary heat with potentiel flow',width=2,**latex_property)
  
  mlab.figure(3)
  simlab.plot(Th,phi)
  mlab.colorbar(title='phi',orientation='vertical')
  simlab.plotiso(Th,phi,contours=15,color='white',line_width=1,opacity=0.5)
  mlab.view(0,0)
  #simlab.plotmesh(Th,d=1,color='Black',line_width=2)
  mlab_latex3D('Velocity potential $\phi$',width=1.2,**latex_property)
  
  mlab.figure(4)
  simlab.plot(Th,v2)
  mlab.colorbar(title='|V|',orientation='vertical')
  simlab.plotiso(Th,v2,contours=15,color='white',line_width=1,opacity=0.5)
  mlab.view(0,0)
  #simlab.plotmesh(Th,d=1,color='Black',line_width=2)
  mlab_latex3D('Norm of the potential flow $\mathbf{V}$',width=2,**latex_property)
  
  mlab.figure(5)
  simlab.quiver(Th,V,scalars=u,colormap='jet',scale_factor=0.0015,line_width=1,mask_points=Th.nq//4000)
  #mlab.vectorbar(title='Velocity')
  mlab.scalarbar(title='u',orientation='vertical')
  mlab.view(0,0)
  simlab.plotmesh(Th,color='LightGray',opacity=0.05)
  simlab.plotmesh(Th,d=1,color='Black',line_width=2)
  mlab_latex3D('Potential flow $\mathbf{V}$ colorized with heat $u$',width=2,**latex_property)
  
  mlab.figure(6)
  simlab.quiver(Th,V,scalars=u,colormap='jet',scale_factor=0.0015,line_width=1,mask_points=Th.nq//4000)
  simlab.plotmesh(Th,color='LightGray',opacity=0.05)
  simlab.plotmesh(Th,d=1,color='Black',line_width=2)
  mlab.view(0.0, 0.0,distance=1.1,focalpoint=(-0.65,0.,0.))

def matplotlib_plot(Th,phi,V,u):
  import matplotlib.pyplot as plt
  import fc_simesh_matplotlib.siMesh as siplt
  from fc_tools.matplotlib import set_axes_equal,DisplayFigures
  v2=np.sqrt(V[0]**2+V[1]**2)
  mv2=np.max(v2)
  
  plt.close('all')
  plt.ion()
  DisplayFigures(nfig=6)

  plt.figure(1)
  siplt.plotmesh(Th,color='LightGray',alpha=0.3)
  siplt.plotmesh(Th,d=1,legend=True)
  set_axes_equal()

  plt.figure(2)
  siplt.plot(Th,u,colormap='jet')
  plt.colorbar(label='$u$')
  siplt.plotiso(Th,u,isorange=np.arange(1,17),color='white',linewidth=1,alpha=0.5)
  plt.title('2D stationary heat with potentiel flow')
  plt.axis('off');set_axes_equal()

  plt.figure(3)
  siplt.plot(Th,phi)
  plt.colorbar(label='$\phi$')
  siplt.plotiso(Th,phi,niso=10,color='white',linewidth=1,alpha=0.5)
  plt.title('Velocity potential $\phi$')
  plt.axis('off');set_axes_equal()
  
  plt.figure(4)
  siplt.plot(Th,v2)
  plt.colorbar(label='$\|V\|_2$')
  siplt.plotiso(Th,v2,niso=15,color='white',linewidth=1,alpha=0.5)
  plt.title('Norm of the potential flow $V$')
  plt.axis('off');set_axes_equal()

  plt.figure(5)
  siplt.quiver(Th,V,scalars=u,scale_factor=2*Th.get_h()/mv2,nvec=Th.nq//2,colormap='jet',linewidth=1)
  plt.colorbar(label='$u$')
  plt.title('Potential flow $\mathbf{V}$ colorized with heat $u$')
  siplt.plotmesh(Th,color='LightGray',alpha=0.3)
  siplt.plotmesh(Th,d=1,color='Black',linewidth=2)
  plt.axis('off');set_axes_equal()
  
  plt.figure(6)
  siplt.quiver(Th,V,scalars=u,scale_factor=2*Th.get_h()/mv2,nvec=Th.nq//2,colormap='jet',linewidth=1)
  siplt.plotmesh(Th,color='LightGray',alpha=0.3)
  siplt.plotmesh(Th,d=1,color='Black',linewidth=2)
  plt.axis('off');set_axes_equal()
  plt.axis([-1, -0.2, -0.22, 0.22])




from fc_simesh import siMesh

from . import operators
from .FEMtools import *

import numpy as np
from scipy import linalg
from scipy import sparse
import itertools
import types
from math import *
import inspect

#def DAssemblyP1_base(Th,D,*args,**kwargs):
def DAssemblyP1_base(Th,Lop,G=None,**kwargs):
  assert(operators.isoperatorL(Lop))
  spformat=kwargs.get("spformat", "csc")
  Lh=Lhoperator(Lop,Th)
  M=sparse.lil_matrix((Th.nq,Th.nq))
  ndfe=Th.d+1;
  for k in range(Th.nme):
    E=DElemP1(Th,Lh,k);
    for il in range(ndfe):
      i=Th.me[k,il]
      for jl in range(ndfe):
        j=Th.me[k,jl]
        M[i,j]=M[i,j]+E[il,jl]
  if (spformat=="lil"):
    return M
  else:
    return eval("M.to%s()"%spformat, globals(), locals())

def DAssemblyP1_basea(Th,Lop,G=None,**kwargs):
  assert(operators.isoperatorL(Lop))
  spformat=kwargs.get("spformat", "csc")
  Lh=Lhoperator(Lop,Th)
  M=sparse.lil_matrix((Th.nq,Th.nq))
  ndfe=Th.d+1;
  for k in range(Th.nme):
    E=DElemP1(Th,Lh,k);
    for il in range(ndfe):
      i=Th.me[k,il]
      for jl in range(ndfe):
        j=Th.me[k,jl]
        M[i,j]=M[i,j]+E[il,jl]
  if (spformat=="lil"):
    return M
  else:
    return eval("M.to%s()"%spformat, globals(), locals())

def HAssemblyP1_base(Th,Hop,Num,**kwargs):
  spformat=kwargs.get("spformat", "csc")
  assert(operators.isoperatorH(Hop))
  assert(isinstance(Th,siMesh))
  m=Hop.m;nq=Th.nq;ndof=m*nq
  M=sparse.csc_matrix((ndof,ndof))
  ndfe=Th.d+1;
  VFNum=getVFindices(Num,m,nq)
  I=np.arange(Th.nq)
  for il in range(m):
    Il=VFNum(I,il)
    for jl in range(m):
      Jl=VFNum(I,jl)
      D=DAssemblyP1_base(Th,Hop.H[il][jl]).tocoo()
      Dr=sparse.csc_matrix((D.data,(VFNum(D.row,il),VFNum(D.col,jl))),shape=(ndof,ndof))
      M=M+Dr
  if (spformat=="csc"):
    return M
  else:
    return eval("M.to%s()"%spformat, globals(), locals())    

def HAssemblyP1_baseb(Th,Hop,Num,**kwargs):
  spformat=kwargs.get("spformat", "csc")
  assert(operators.isoperatorH(Hop))
  assert(isinstance(Th,siMesh))
  m=Hop.m;nq=Th.nq;ndof=m*nq
  M=sparse.lil_matrix((ndof,ndof))
  ndfe=Th.d+1;
  VFNum=getVFindices(Num,m,nq)
  I=np.arange(Th.nq)
  for il in range(m):
    Il=VFNum(I,il)
    for jl in range(m):
      Jl=VFNum(I,jl)
      M[np.ix_(Il,Jl)]=DAssemblyP1_base(Th,Hop.H[il][jl],spformat="lil")
  if (spformat=="lil"):
    return M
  else:
    return eval("M.to%s()"%spformat, globals(), locals())    

def HAssemblyP1_basea(Th,Hop,Num,**kwargs):
  spformat=kwargs.get("spformat", "csc")
  assert(operators.isoperatorH(Hop))
  assert(isinstance(Th,siMesh))
  m=Hop.m;nq=Th.nq;ndof=m*nq
  M=sparse.lil_matrix((ndof,ndof))
  ndfe=Th.d+1;
  VFNum=getVFindices(Num,m,nq)
  I=np.arange(Th.nq)
  for jl in range(m):
    Jl=VFNum(I,jl)
    for il in range(m):
      Il=VFNum(I,il)
      M[np.ix_(Il,Jl)]=DAssemblyP1_base(Th,Hop.H[il][jl],spformat="lil")
  if (spformat=="lil"):
    return M
  else:
    return eval("M.to%s()"%spformat, globals(), locals())    

def DElemP1(Th,Dh,k):
  d=Th.d
  vol=Th.vols[k];
  mek=Th.me[k]
  #E=np.zeros((d+1,d+1))
  if (Dh.a0!=None):
    E=DElemP1_guv(d,vol,getLocalfData(Dh.a0,Th,k))
  else:
    E=np.zeros((d+1,d+1))
  if (Dh.order==0):
    return E
  G=GradientLocal(Th.q[Th.me[k]],Th.vols[k])
 
  if (Dh.A!=None):
    for i in range(d):
      for j in range(d):
        if (Dh.A[i][j]!=None):
          E=DElemP1_gdudv(d,vol,getLocalfData(Dh.A[i][j],Th,k),j,i,G,E)
          
  if (Dh.b!=None):
    for i in range(d):
      if (Dh.b[i]!=None):
        E=DElemP1_gudv(d,vol,getLocalfData(Dh.b[i],Th,k),i,G,E)
  if Dh.c!=None:
    for i in range(d):
      if (Dh.c[i]!=None):
        E=DElemP1_gduv(d,vol,getLocalfData(Dh.c[i],Th,k),i,G,E)
  return E

def DElemP1_guv(d,vol,gl):
  MatElem=lambda il,jl: (factorial(d)/float(factorial(d+3)))*(1+(il==jl))*vol*(gl.sum()+gl[il]+gl[jl])  
  De=np.zeros((d+1,d+1))
  for il in range(d+1):
    for jl in range(d+1):
      De[il,jl]=MatElem(il,jl)
  return De

def DElemP1_gdudv(d,vol,gl,i,j,G,De):
  MatElem=lambda il,jl: (factorial(d)/float(factorial(d+1)))*vol*gl.sum()*G[i,jl]*G[j,il]  
  for il in range(d+1):
    for jl in range(d+1):
      De[il,jl]+=MatElem(il,jl)
  return De
      
def DElemP1_gduv(d,vol,gl,i,G,De):
  MatElem=lambda il,jl: (factorial(d)/float(factorial(d+2)))*vol*(gl.sum()+gl[il])*G[i,jl]  
  for il in range(d+1):
    for jl in range(d+1):
      De[il,jl]+=MatElem(il,jl)
  return De

def DElemP1_gudv(d,vol,gl,i,G,De):
  MatElem=lambda il,jl: (factorial(d)/float(factorial(d+2)))*vol*(gl.sum()+gl[jl])*G[i,il]  
  for il in range(d+1):
    for jl in range(d+1):
      De[il,jl]-=MatElem(il,jl)
  return De

def getLocalfData(fh,Th,k):
  return fh[Th.me[k]]

def GradientLocal(ql,vol):
  d=ql.shape[1];
  G=np.zeros((d,d+1))
  if (d==1):
    G[0,0]=-1/vol;
    G[0,1]=1/vol;      
    return G;
  if (d==2):
    D23=ql[1]-ql[2];D13=ql[0]-ql[2];D12=ql[0]-ql[1];
    C=1/(2.*vol)
    G[0,0]=D23[1]*C;G[1,0]=-D23[0]*C
    G[0,1]=-D13[1]*C;G[1,1]=D13[0]*C
    G[0,2]=D12[1]*C;G[1,2]=-D12[0]*C
    return G;
  if (d==3):
    D12=ql[0]-ql[1];D13=ql[0]-ql[2];D14=ql[0]-ql[3]
    D23=ql[1]-ql[2];D24=ql[1]-ql[3];D34=ql[2]-ql[3]
    C=1/(6*vol)
    G[0,0]=(-D23[1]*D24[2] + D23[2]*D24[1])*C
    G[1,0]=( D23[0]*D24[2] - D23[2]*D24[0])*C
    G[2,0]=(-D23[0]*D24[1] + D23[1]*D24[0])*C
    G[0,1]=( D13[1]*D14[2] - D13[2]*D14[1])*C
    G[1,1]=(-D13[0]*D14[2] + D13[2]*D14[0])*C
    G[2,1]=( D13[0]*D14[1] - D13[1]*D14[0])*C
    G[0,2]=(-D12[1]*D14[2] + D12[2]*D14[1])*C
    G[1,2]=( D12[0]*D14[2] - D12[2]*D14[0])*C
    G[2,2]=(-D12[0]*D14[1] + D12[1]*D14[0])*C
    G[0,3]=( D12[1]*D13[2] - D12[2]*D13[1])*C
    G[1,3]=(-D12[0]*D13[2] + D12[2]*D13[0])*C
    G[2,3]=( D12[0]*D13[1] - D12[1]*D13[0])*C
    return G
  return ComputeGradientdD(ql,vol)  

def ComputeGradientdD(ql,vol):
  d=ql.shape[1]
  Grad=np.zeros((d+1,d))
  Grad[1:d+1]=np.eye(d);Grad[0]=-1
  X=np.zeros((d,d))
  for i in range(d):
    X[:,i]=ql[i+1]-ql[0]   
  G1=np.dot(Grad,linalg.inv(X))
  return G1.transpose()

# Operator L discretized on mesh Th
class Lhoperator:
  def __init__(self,Lop,Th):
    assert(operators.isoperatorL(Lop))
    assert(isinstance(Th,siMesh) or isinstance(Th,mesh.bdMesh))
    self.d=Lop.d
    self.m=1
    self.order=Lop.order
    if (Lop.a0 == None):
      self.a0=None
    else:
      self.a0=setFdata(Lop.a0,Th)
    
    if (Lop.A == None):
      self.A=None
    else:
      self.A = operators.NoneMatrix(self.d,self.d)
      for i in range(self.d):
        for j in range(self.d):
          self.A[i][j]=setFdata(Lop.A[i][j],Th)
          
    if (Lop.b == None):
      self.b=None
    else:
      self.b = operators.NoneVector(self.d)
      for i in range(self.d):
        self.b[i]=setFdata(Lop.b[i],Th)
    if (Lop.c == None):
      self.c=None
    else:
      self.c = operators.NoneVector(self.d)
      for i in range(self.d):
        self.c[i]=setFdata(Lop.c[i],Th)

# Operator Hop discretized on mesh Th
class Hhoperator:
  def __init__(self,Hop,Th):
    assert(operators.isoperatorH(Hop))
    assert(isinstance(Th,siMesh))
    self.d=Hop.d
    self.m=Hop.m
    self.H=operators.NoneMatrix(self.m,self.m)
    for i in range(self.m):
      for j in range(self.m):
        self.H[i][j]=Lhoperator(Hop.H[i][j],Th)


import os
import fc_simesh

def get_geodirs(dim,d):
  s=get_pathname(dim,d)
  assert s is not None,"Unable to find geo directory for dim=%d and d=%d"%(dim,d)
  fullname=os.path.dirname(os.path.abspath(__file__))
  loc_dir=fullname+os.sep+'geodir'+os.sep+s
  from fc_simesh import sys as fsys
  simesh_dir=fsys.get_geodirs(dim,d)
  simesh_dir.insert(0,loc_dir)
  return simesh_dir
  
def get_pathname(dim,d):
  if dim==2 and d==2:
    return '2d'
  if dim==3 and d==2:
    return '3ds'
  if dim==3 and d==3:
    return '3d'
  return None

# geofile without path and without ext
def get_geo(dim,d,geofile):
  (geodir,geofile)=os.path.split(geofile)
  #assert geodir=='', "geofile must be without path: %s"%(geofile)
  (geofile,geoext)=os.path.splitext(geofile)
  assert geoext=='.geo' or geoext=='', "Extension must be .geo in %s"%geofile
  if geodir!='':
    fullgeo=geodir+os.sep+geofile+'.geo'
    if os.path.isfile(fullgeo):
      return (geodir,geofile)
    print('geofile %s.geo not found in directory %s'%(geofile,geodir))
    print('Trying default directories ...')
  geodirs=get_geodirs(dim,d)
  for di in geodirs:
    fullgeo=di+os.sep+geofile+'.geo'
    if os.path.isfile(fullgeo):
      return (di,geofile)
  assert False,"geofile %s.geo not found!"%geofile


import numpy as np
from scipy import sparse

# A(tgr,tgc) <- A(tgr,tgc)+Ai
def sum_sub_mat(A,Ai,tgr,tgc):
  if Ai is None:
    return A
  if A.shape==Ai.shape:
    return A+Ai
  (I,J,K)=sparse.find(Ai)
  A+=sparse.csc_matrix((K,(tgr[I],tgc[J])),shape=A.shape)
  return A

def csr_zero_rows(A, rows_to_zero):
  if not isinstance(A, sparse.csc_matrix):
    raise ValueError('Matrix given must be of CSR format.')
  mask=np.in1d(A.indices, rows_to_zero)
  A.data[mask]=0
  #A.eliminate_zeros()
  #A.eliminate_zeros()

def csr_one_rows(A, rows_to_zero):
  if not isinstance(A, sparse.csc_matrix):
    raise ValueError('Matrix given must be of CSR format.')
  mask=np.in1d(A.indices, rows_to_zero)
  A.data[mask]=1
  
def csr_sum_onerow(A, row):
  if not isinstance(A, sparse.csc_matrix):
    raise ValueError('Matrix given must be of CSR format.')
  mask=np.in1d(A.indices, row)
  return sum(A.data[mask])


import numpy as np

from fc_vfemp1.operators import Loperator
from fc_vfemp1.examples import set_examples
from fc_vfemp1.examples.set_examples import *
from fc_vfemp1 import FEM
from fc_tools.others import isModuleFound

def BVPElectrostatic2D01(**kwargs):
  N=kwargs.pop('N',80)
  kwargs['sigma1']=kwargs.get('sigma1',1)
  kwargs['sigma2']=kwargs.get('sigma2',10)
  isMayavi=kwargs.pop('mayavi',isModuleFound('fc_simesh_mayavi'))
  isMatplotlib=kwargs.pop('matplotlib',isModuleFound('fc_simesh_matplotlib'))
  bvp,info=set_examples.setBVPElectrostatic2D01(N,**kwargs)
  u=bvp.solve()
  Th=bvp.Th
  E=np.zeros((2,Th.nq))
  Lop=Loperator(dim=2,d=2,c=[1,0])
  E[0]=FEM.Apply(Lop,Th,u)#,solver=solver,perm=perm)
  Lop=Loperator(dim=2,d=2,c=[0,1])
  E[1]=FEM.Apply(Lop,Th,u)#,solver=solver,perm=perm)
  ENorm=np.sqrt(E[0]**2+E[1]**2)
   
  def BVPElectrostatic2D01_mayavi(Th,u,E,ENorm):
    from mayavi import mlab
    import fc_simesh_mayavi.siMesh as simlab
    if len(mlab.get_engine().scenes)>0:
      mlab.close(all=True)

    mlab.figure(1)
    simlab.plotmesh(Th,legend=True)
    mlab.view(0,0)

    mlab.figure(2)
    simlab.plotmesh(Th,d=1,legend=True,line_width=2)
    simlab.plotmesh(Th,color='LightGray',opacity=0.3)
    mlab.view(0,0)

    mlab.figure(3)
    simlab.plot(Th,u,colormap='viridis')
    simlab.plotiso(Th,u,contours=25,color='white',line_width=1)
    simlab.plotmesh(Th,d=1,color='black',line_width=1.5)
    mlab.view(0,0)
    mlab.colorbar(title='u',orientation='vertical')

    mlab.figure(4)
    simlab.plotiso(Th,u,contours=25,colormap='viridis')
    mlab.view(0,0)
    mlab.colorbar(title='u',orientation='vertical')
    simlab.plotmesh(Th,color='LightGray',opacity=0.05)
    simlab.plotmesh(Th,d=1,color='black',line_width=1.5)

    mlab.figure(5)
    simlab.plot(Th,ENorm,d=2,colormap='jet')
    simlab.plotiso(Th,ENorm,d=2,contours=20,color=(1,1,1),line_width=1)
    simlab.plotmesh(Th,d=1,color='black',line_width=1.5)
    mlab.view(0,0)
    mlab.colorbar(title='||E||',orientation='vertical')
    
    mlab.figure(6)
    simlab.plotiso(Th,ENorm,contours=25,colormap='jet')
    mlab.view(0,0)
    mlab.colorbar(title='||E||',orientation='vertical')
    simlab.plotmesh(Th,color='LightGray',opacity=0.05)
    simlab.plotmesh(Th,d=1,color='black',line_width=1.5)

    mlab.figure(7)
    simlab.quiver(Th,E,scale_factor=0.3e-3)
    mlab.colorbar(title='E',orientation='vertical')
    simlab.plotmesh(Th,d=1,color='black',line_width=1.5)
    mlab.view(0,0)

  def BVPElectrostatic2D01_matplotlib(Th,u,E,ENorm):
    import matplotlib.pyplot as plt
    import fc_simesh_matplotlib.siMesh as siplt
    from fc_tools.matplotlib import set_axes_equal,DisplayFigures
    plt.close('all')
    plt.ion()
    DisplayFigures(nfig=6)
    plt.figure(1)
    siplt.plotmesh(Th,legend=True)
    set_axes_equal()
    
    plt.figure(2)
    siplt.plotmesh(Th,d=1,legend=True,linewidth=2)
    siplt.plotmesh(Th,color='LightGray',alpha=0.3)
    set_axes_equal()

    plt.figure(3)
    siplt.plot(Th,u,colormap='viridis')
    plt.colorbar(label='u')
    siplt.plotiso(Th,u,contours=25,color='white',linewidth=2)
    plt.axis('off');set_axes_equal()

    plt.figure(4)
    siplt.plotiso(Th,u,contours=25,colormap='viridis')
    plt.colorbar(label='u')
    siplt.plotmesh(Th,color='LightGray',alpha=0.05)
    siplt.plotmesh(Th,d=1,color='black',linewidth=1.5)
    plt.axis('off');set_axes_equal()

    plt.figure(5)
    siplt.plot(Th,ENorm,d=2,colormap='jet')
    plt.colorbar(label='$\|E\|$ ')
    siplt.plotiso(Th,ENorm,d=2,contours=20,color=(1,1,1),linewidth=1)
    siplt.plotmesh(Th,d=1,color='black',linewidth=1.5)
    plt.axis('off');set_axes_equal()

    plt.figure(6)
    siplt.quiver(Th,E,scale_factor=0.3e-3)
    plt.colorbar(label='$E$ ')
    siplt.plotmesh(Th,d=1,color='black',linewidth=1.5)
    plt.axis('off');set_axes_equal()
    
  if isMayavi:
    BVPElectrostatic2D01_mayavi(Th,u,E,ENorm)
  if isMatplotlib:
    BVPElectrostatic2D01_matplotlib(Th,u,E,ENorm)
  return {'bvp':bvp,'u':u,'E':E}

def BVPPoisson2D(**kwargs):
  kwargs['verbose']=kwargs.get('verbose',2)
  version=kwargs.pop('version','ex01')
  N=kwargs.pop('N',50)
  isMayavi=kwargs.pop('mayavi',isModuleFound('fc_simesh_mayavi'))
  isMatplotlib=kwargs.pop('matplotlib',isModuleFound('fc_simesh_matplotlib'))
  bvp,info=globals()['setBVPPoisson2D_'+version](N,**kwargs)
  verbose=kwargs.get('verbose',1)
  if verbose>0:
    print("*** Solving BVP")
  u,err,errL2,errH1=SolvePoisson2Dex(bvp,info)
  if verbose>0:
    print("  Error L2: %e, Error H1: %e\n"%(errL2,errH1))
  Th=bvp.Th  
  if isMayavi:
    BVPPoisson2D_mayavi(Th,u,err)
  if isMatplotlib:
    BVPPoisson2D_matplotlib(Th,u,err)
  return {'bvp':bvp,'info':info,'u':u,'errH1':errH1,'errL2':errL2}
    
def SolvePoisson2Dex(bvp,info):
  u=bvp.solve()
  Th=bvp.Th
  uex=info['solex']
  Uex=Th.feval(uex)
  err=np.abs(u-Uex)
  errH1,M,K=FEM.NormH1(Th,err)
  errH1=errH1/FEM.NormH1(Th,u,Mass=M,Stiff=K)[0]
  errL2=FEM.NormL2(Th,err,Mass=M)[0]/FEM.NormL2(Th,u,Mass=M)[0]
  return u,err,errL2,errH1
    
def BVPPoisson2D_mayavi(Th,u,err):
  from mayavi import mlab
  import fc_simesh_mayavi.siMesh as simlab
  if len(mlab.get_engine().scenes)>0:
    mlab.close(all=True)
  mlab.figure(1)
  simlab.plot(Th,u)
  mlab.view(0,0)
  mlab.colorbar(title='u',orientation='vertical')

  mlab.figure(2)
  simlab.plotiso(Th,u,contours=15)
  mlab.view(0,0)
  mlab.colorbar(title='u',orientation='vertical')
  simlab.plotmesh(Th,color='LightGray',opacity=0.05)

  mlab.figure(3)
  simlab.plot(Th,err,colormap='jet')
  mlab.view(0,0)
  mlab.colorbar(title='error',orientation='vertical')

  mlab.figure(4)
  simlab.plotiso(Th,err,contours=15,colormap='jet')
  mlab.view(0,0)
  t=mlab.colorbar(title='error',orientation='vertical')
  t.label_text_property.font_size=16 # ??? don't work
  simlab.plotmesh(Th,color='LightGray',opacity=0.05)
  
def BVPPoisson2D_matplotlib(Th,u,err):
  import matplotlib.pyplot as plt
  import fc_simesh_matplotlib.siMesh as siplt
  from fc_tools.matplotlib import set_axes_equal,DisplayFigures
  plt.close('all')
  plt.ion()
  DisplayFigures(nfig=4)
  plt.figure(1)
  siplt.plot(Th,u)
  plt.colorbar(label='u')
  set_axes_equal()

  plt.figure(2)
  siplt.plotiso(Th,u,contours=15)
  plt.colorbar(label='u')
  siplt.plotmesh(Th,color='LightGray',alpha=0.3)
  plt.axis('off');set_axes_equal()

  plt.figure(3)
  siplt.plot(Th,err,colormap='jet')
  plt.colorbar(label='error')
  plt.axis('off');set_axes_equal()

  plt.figure(4)
  siplt.plotiso(Th,err,contours=15,colormap='jet')
  plt.colorbar(label='error')
  siplt.plotmesh(Th,color='LightGray',alpha=0.3)
  plt.axis('off');set_axes_equal()
  plt.show()


import numpy as np

from fc_vfemp1.operators import Loperator
from fc_vfemp1.examples.set_examples import setBVPCondenser2D01
from fc_vfemp1 import FEM
from fc_tools.others import isModuleFound

bvp,info=setBVPCondenser2D01(20,verbose=2)
u=bvp.solve()

Th=bvp.Th
if isModuleFound('fc_simesh_mayavi'):
  from mayavi import mlab
  import fc_simesh_mayavi.siMesh as simlab
  if len(mlab.get_engine().scenes)>0:
    mlab.close(all=True)

  mlab.figure(1)
  simlab.plotmesh(Th,legend=True)
  mlab.view(0,0)

  mlab.figure(2)
  simlab.plotmesh(Th,d=1,legend=True,line_width=2)
  simlab.plotmesh(Th,color='LightGray',opacity=0.3)
  mlab.view(0,0)

  mlab.figure(3)
  simlab.plot(Th,u,plane=False)
  simlab.plotiso(Th,u,contours=25,color='white',line_width=2,plane=False)
  #mlab.view(0,0)
  cb=mlab.colorbar(title='u',orientation='vertical')
  #cb.lut_mode='viridis'

  mlab.figure(4)
  simlab.plotiso(Th,u,contours=25)
  mlab.view(0,0)
  mlab.colorbar(title='u',orientation='vertical')
  simlab.plotmesh(Th,color='LightGray',opacity=0.05)
  simlab.plotmesh(Th,d=1,color='black',line_width=1.5)

if isModuleFound('fc_simesh_matplotlib'):  
  import matplotlib.pyplot as plt
  import fc_simesh_matplotlib.siMesh as siplt
  from fc_tools.matplotlib import set_axes_equal,DisplayFigures
  plt.close('all')
  plt.ion()
  plt.figure(1)
  siplt.plotmesh(Th,legend=True)
  set_axes_equal()
  
  plt.figure(2)
  siplt.plotmesh(Th,d=1,legend=True,linewidth=2)
  siplt.plotmesh(Th,color='LightGray',alpha=0.3)
  set_axes_equal()

  plt.figure(3)
  hp=siplt.plot(Th,u,plane=False)
  plt.colorbar(hp,label='u')
  # siplt.plotiso(Th,u,contours=25,color='white',linewidth=2,plane=False)
  plt.axis('off')#;set_axes_equal()

  plt.figure(4)
  pi=siplt.plotiso(Th,u,contours=25)
  plt.colorbar(pi,label='u')
  siplt.plotmesh(Th,color='LightGray',alpha=0.05)
  siplt.plotmesh(Th,d=1,color='black',linewidth=1.5)
  plt.axis('off');set_axes_equal()

  DisplayFigures()



from math import factorial
from scipy import sparse
import numpy as np

from fc_simesh.siMeshElt import siMeshElt

from . import FEMtools
from . import operators

#def DAssemblyP1_OptV3(sTh,Lop,**kwargs):
  #dtype=kwargs.get('dtype',float)
  #Kg=KgP1_OptV3(sTh,Lop,dtype=dtype)
  #Ig,Jg=IgJgP1_OptV3(sTh.d,sTh.nme,sTh.me)
  #N=sTh.nme*(sTh.d+1)**2
  #A=sparse.csc_matrix((np.reshape(Kg,N),(np.reshape(Ig,N),np.reshape(Jg,N))),shape=(sTh.nq,sTh.nq))
  #A.eliminate_zeros()
  #return A

#def DAssemblyP1_OptV3new(sTh,Lop,**kwargs):
def DAssemblyP1_OptV3(sTh,Lop,**kwargs):
  assert(isinstance(sTh, siMeshElt))
  assert(isinstance(Lop, operators.Loperator))
  dtype=kwargs.get('dtype',float)
  local=kwargs.get('local',False)
  Kg=KgP1_OptV3(sTh,Lop,dtype=dtype)
  if local:
    ndof=sTh.nq
    Ig,Jg=IgJgP1_OptV3(sTh.d,sTh.nme,sTh.me)
  else:
    ndof=sTh.nqParent
    Ig,Jg=IgJgP1_OptV3(sTh.d,sTh.nme,sTh.toParent[sTh.me])
  N=sTh.nme*(sTh.d+1)**2
  A=sparse.csc_matrix((np.reshape(Kg,N),(np.reshape(Ig,N),np.reshape(Jg,N))),shape=(ndof,ndof))
  A.eliminate_zeros()
  return A
  
def HAssemblyP1_OptV3(sTh,Hop,**kwargs):
  assert(isinstance(sTh, siMeshElt))
  assert(isinstance(Hop, operators.Hoperator))
  dtype=kwargs.get('dtype',float)
  spformat=kwargs.get("spformat", "csc")
  local=kwargs.get('local',True)
  Num=kwargs.get('Num',1)
  block=kwargs.get('block',True)
  
  if local:
    nq=sTh.nq
    Ig,Jg=IgJgP1_OptV3(sTh.d,sTh.nme,sTh.me)
  else:
    nq=sTh.nqGlobal
    Ig,Jg=IgJgP1_OptV3(sTh.d,sTh.nme,sTh.toGlobal[sTh.me])
  N=sTh.nme*(sTh.d+1)**2
  Ig=np.reshape(Ig,N)
  Jg=np.reshape(Jg,N)
  
  m=Hop.m;ndof=m*nq
  VFInd=FEMtools.getVFindices(Num,m,nq) 
  if block:
    #M=m*[m*[None]] # Failed all lines points to the same line
    M=FEMtools.NoneMatrix(m,m)
  else:
    M=sparse.csc_matrix((ndof,ndof),dtype=dtype)
  
  for i in range(m):
    for j in range(m):
      if Hop.H[i][j] is not None:
        Kg=np.reshape(KgP1_OptV3(sTh,Hop.H[i][j],dtype=dtype),N)
        if block:
          M[i][j]=sparse.csc_matrix((Kg,(Ig,Jg)),shape=(nq,nq),dtype=dtype)
        else:  
          M=M+sparse.csc_matrix((Kg,(VFNum(Ig,i),VFNum(Jg,j))),shape=(ndof,ndof),dtype=dtype)
      #print('*** i=%d,j=%d'%(i,j))
      #print(M)
  if (spformat=="csc"):
    return M
  else:
    return eval("M.to%s()"%spformat, globals(), locals())
  
def RHSAssembly(sTh,f,**kwargs):
  assert(isinstance(sTh, siMeshElt))
  dtype=kwargs.get('dtype',float)
  #spformat=kwargs.get("spformat", "csc")
  local=kwargs.get('local',True)
  Num=kwargs.get('Num',1)
  block=kwargs.get('block',True)
  m=kwargs.get('m',1)
  if local:
    nq=sTh.nq
  else:
    nq=sTh.nqGlobal
    
  #N=sTh.nme*(sTh.d+1)**2
  ndof=m*nq
  #VFInd=FEMtools.getVFindices(Num,m,nq) 
  if block:
    b=FEMtools.NoneVector(m)
  else:
    b=np.zeros((ndof,1))
  if f is None:
    return b
  if (m==1) and not (isinstance(f,list)):
    f=[f]
  VFInd=FEMtools.getVFindices(Num,m,nq)   
  Mop=operators.Loperator(dim=sTh.dim,d=sTh.d,a0=1)
  M=DAssemblyP1_OptV3(sTh,Mop,local=True)
  if not block:
    I=np.arange(sTh.nq)
  for i in range(m):
    if f[i] is not None:
      fh=sTh.feval(f[i]).reshape((sTh.nq,1))
      Fh=M*fh
      if block:
        b[i]=Fh
      else:
        b[VFInd(I,i)]=Fh
  return b

def IgJgP1_OptV3(d,nme,me):
  ndfe=d+1
  Ig=np.zeros((nme,ndfe,ndfe),dtype=np.int32)
  Jg=np.zeros((nme,ndfe,ndfe),dtype=np.int32)
  for il in range(ndfe):
    mel=me[il]
    for jl in range(ndfe):
      Ig[:,il,jl]=Jg[:,jl,il]=mel
  return Ig,Jg
  
def KgP1_OptV3(sTh,Lop,**kwargs):
  d=sTh.d;ndfe=d+1
  dtype=kwargs.get('dtype', float)
  Kg=np.zeros((sTh.nme,ndfe,ndfe),dtype)
  Kg=Kg+KgP1_OptV3_guv(sTh,Lop.a0,dtype)
  if (Lop.get_order()==0):
    return Kg
  if Lop.A is not None:
    for i in range(d):
      for j in range(d):
          Kg=Kg+KgP1_OptV3_gdudv(sTh,Lop.A[i][j],j,i,dtype)
  if Lop.b is not None:
    for i in range(d):
        Kg=Kg-KgP1_OptV3_gudv(sTh,Lop.b[i],i,dtype)
  if Lop.c is not None:
    for i in range(d):
        Kg=Kg+KgP1_OptV3_gduv(sTh,Lop.c[i],i,dtype)
  return Kg
  
def KgP1_OptV3_guv(sTh,g,dtype):
  if g is None: return 0
  gh=FEMtools.setFdata(g,sTh,dtype=dtype)
  d=sTh.d;ndfe=d+1
  Kg=np.zeros((sTh.nme,ndfe,ndfe),dtype=dtype)
  gme=gh[sTh.me]
  gs=(gme.sum(axis=0)).reshape(sTh.nme)
  KgElem=lambda il,jl: (factorial(d)/float(factorial(d+3)))*(1+(il==jl))*sTh.vols*(gs+gme[il]+gme[jl])
  for il in range(ndfe):
    for jl in range(ndfe):
      Kg[:,il,jl]=KgElem(il,jl)
  return Kg

def KgP1_OptV3_gdudv(sTh,g,i,j,dtype):
  if g is None: return 0
  G=sTh.GradBaCo
  gh=FEMtools.setFdata(g,sTh,dtype=dtype)
  d=sTh.d;ndfe=d+1
  Kg=np.zeros((sTh.nme,ndfe,ndfe),dtype=dtype)
  gme=gh[sTh.me]
  gs=(gme.sum(axis=0)).reshape(sTh.nme)
  KgElem=lambda il,jl: (factorial(d)/float(factorial(d+1)))*sTh.vols*gs*G[:,jl,i]*G[:,il,j]
  for il in range(ndfe):
    for jl in range(ndfe):
      Kg[:,il,jl]=KgElem(il,jl)
  return Kg
  
def KgP1_OptV3_gudv(sTh,g,i,dtype):
  if g is None: 
    return 0
  G=sTh.GradBaCo
  gh=FEMtools.setFdata(g,sTh,dtype=dtype)
  d=sTh.d;ndfe=d+1
  Kg=np.zeros((sTh.nme,ndfe,ndfe),dtype=dtype)
  gme=gh[sTh.me]
  gs=(gme.sum(axis=0)).reshape(sTh.nme)
  KgElem=lambda il,jl: (factorial(d)/float(factorial(d+2)))*sTh.vols*(gs+gme[jl])*G[:,il,i]
  for il in range(ndfe):
    for jl in range(ndfe):
      Kg[:,il,jl]=KgElem(il,jl)
  return Kg
  
def KgP1_OptV3_gduv(sTh,g,i,dtype):
  if g is None: 
    return 0
  G=sTh.GradBaCo
  gh=FEMtools.setFdata(g,sTh,dtype=dtype)
  d=sTh.d;ndfe=d+1
  Kg=np.zeros((sTh.nme,ndfe,ndfe),dtype=dtype)
  gme=gh[sTh.me]
  gs=(gme.sum(axis=0)).reshape(sTh.nme)
  KgElem=lambda il,jl: (factorial(d)/float(factorial(d+2)))*sTh.vols*(gs+gme[il])*G[:,jl,i]
  for il in range(ndfe):
    for jl in range(ndfe):
      Kg[:,il,jl]=KgElem(il,jl)
  return Kg



__version__ = '0.0.2'

#from .common import *
#from .operators import *
#from .pde import *
#from .FEMOptV3 import *
#from .FEMbase import *
#from .FEMtools import *
#from .FEM import *
#from .samples import *




from setuptools import setup, find_packages

setup_defaults = {  
   'name'        : 'fc_vfemp1',
   'description' : 'Solve scalar and vector Boundary Value Problems by P1-Lagrange ﬁnite element method in any space dimension',
   'version'     : '0.0.2',
   'url'         : 'http://www.math.univ-paris13.fr/~cuvelier/software',
   'author'      : 'Francois Cuvelier',
   'author_email': 'cuvelier@math.univ-paris13.fr',
   'license'     : 'BSD',
   'packages'    : ['fc_vfemp1'],
   'classifiers':['Topic :: Scientific/Engineering'],
   } 

import glob
geofiles=glob.glob("src/fc_oogmsh/geodir/2d/*.geo")+glob.glob("src/fc_oogmsh/geodir/3d/*.geo")+glob.glob("src/fc_oogmsh/geodir/3ds/*.geo")

setup(name=setup_defaults['name'],
      description = setup_defaults['description'],
      version=setup_defaults['version'],
      url=setup_defaults['url'],
      author=setup_defaults['author'],
      author_email=setup_defaults['author_email'],
      license = setup_defaults['license'],
      platforms=["Linux"],#, "Mac OS-X", 'Windows'],
      #packages=setup_defaults['packages'],
      packages=find_packages('src'),  # include all packages under src
      package_dir={'':'src'},   # tell distutils packages are under src
      classifiers=setup_defaults['classifiers'],
      install_requires=['fc_simesh >= 0.0.5'],
      package_data={
        'fc_vfemp1': ['geodir/2d/*.geo','geodir/3d/*.geo','geodir/3ds/*.geo'],
      },
      extras_require={
        'mayavi':  ["fc_simesh_mayavi"],
        'matplotlib': ["fc_simesh_matplotlib"],
      }
     )


import numpy as np
from fc_vfemp1.operators import Loperator,Hoperator
from fc_vfemp1.BVP import BVP,PDE
from fc_simesh.siMesh import HyperCube
from fc_tools.others import isModuleFound

from numpy import cos,sin
E=21e5;nu=0.45;L=20;N=10
mu= E/(2*(1+nu))
lam = E*nu/((1+nu)*(1-2*nu))
gam=lam+2*mu
uex=[lambda x,y: cos(2*x+y), lambda x,y: sin(x-3*y)]
f=[lambda x,y: 4*lam*cos(2*x + y) + 9*mu*cos(2*x + y) + 3*lam*sin(-x + 3*y) + 3*mu*sin(-x + 3*y) ,
    lambda x,y: 2*lam*cos(2*x + y) + 2*mu*cos(2*x + y) - 9*lam*sin(-x + 3*y) - 19*mu*sin(-x + 3*y)]
g2=[lambda x,y: -3*lam*cos(-x + 3*y) - 2*lam*sin(2*x + y) - 4*mu*sin(2*x + y), 
    lambda x,y: mu*(cos(-x + 3*y) - sin(2*x + y))]
print('------------------------------------------------------') 
print('      2D elasticity BVP on [0,%g]x[-1,1]'%L)
print('         on transform hypercube mesh\n')
print('             -div(sigma(u)) = f ')
print(' and boundary conditions :')
print('   * Dirichlet on [1]')
print('       u_1 = u_2 = 0 on [1],')
print('   * Neumann on [2,3,4]')
print('       sigma(u).n = 0 on [2,3,4]')
print(' with f=(0,-1)^t, lambda=%g, mu=%g\n'%(lam,mu));
print('------------------------------------------------------')
print('1. Setting the mesh using HyperCube function')
mapping=lambda q: np.array([L*q[0],-1+2*q[1]])
Th=HyperCube(2,[2*N,20*N],mapping=mapping)
print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
print('2.a Setting 2D Elasticity B.V.P. problem')

Hop=Hoperator(dim=2,d=2,m=2)
Hop.H[0][0]=Loperator(d=2,A=[[gam,None],[None,mu]]) 
Hop.H[0][1]=Loperator(d=2,A=[[None,lam],[mu,None]]) 
Hop.H[1][0]=Loperator(d=2,A=[[None,mu],[lam,None]]) 
Hop.H[1][1]=Loperator(d=2,A=[[mu,None],[None,gam]])
pde=PDE(Op=Hop,f=[0,-1])
bvp=BVP(Th,pde=pde)
bvp.setDirichlet(1,[0,0])

print('2.b Solving 2D elasticity BVP')
U=bvp.solve(split=True)

Th=bvp.Th

normU=np.sqrt(U[0]**2+U[1]**2)

if isModuleFound('fc_simesh_mayavi'):
  from mayavi import mlab
  import fc_simesh_mayavi.siMesh as simlab
  scale=50;
  if len(mlab.get_engine().scenes)>0:
    mlab.close(all=True)

  mlab.figure(1)
  simlab.plotmesh(Th,color='LightGray',opacity=0.01)
  simlab.plotmesh(Th,d=1,legend=True,line_width=2)
  mlab.view(0,0)

  mlab.figure(2)
  simlab.plotmesh(Th,color='blue',opacity=0.01)
  simlab.plotmesh(Th,color='red',move=[scale*U[0],scale*U[1]])
  mlab.view(0,0)
  
  mlab.figure(3)
  simlab.plot(Th,normU)
  simlab.plotiso(Th,normU,contours=10,color='White',line_width=1)
  simlab.plotmesh(Th,d=1,color='black')
  mlab.view(0,0)
  mlab.colorbar(title='|U|')

if isModuleFound('fc_simesh_matplotlib'):  
  import matplotlib.pyplot as plt
  import fc_simesh_matplotlib.siMesh as siplt
  from fc_tools.matplotlib import set_axes_equal,DisplayFigures
  plt.close('all')
  plt.ion()
  DisplayFigures(nfig=3)
  plt.figure(1)
  siplt.plotmesh(Th,color='LightGray',alpha=0.05)
  siplt.plotmesh(Th,d=1,legend=True,linewidth=2)
  set_axes_equal()
  
  plt.figure(2)
  siplt.plotmesh(Th,color='blue',alpha=0.05)
  siplt.plotmesh(Th,color='red',move=[scale*U[0],scale*U[1]],alpha=0.1)
  set_axes_equal()
  
  plt.figure(3)
  siplt.plot(Th,normU)
  plt.colorbar(label='|u|')
  siplt.plotiso(Th,normU,contours=15,color='white',linewidth=2)
  siplt.plotmesh(Th,d=1,color='black')
  plt.axis('off');set_axes_equal()
  


# Run with: from fc_vfemp1.samples import BVPStationaryConvectionDiffusion2D01
import numpy as np
from fc_vfemp1.examples.set_examples import setBVPStationaryConvectionDiffusion2D01
bvp,info=setBVPStationaryConvectionDiffusion2D01(120,verbose=2)
u=bvp.solve()


from fc_tools.graphics import isMayavi,isMatplotlib
Th=bvp.Th
if isMayavi():
  from mayavi import mlab
  import fc_simesh_mayavi.siMesh as simlab
  if len(mlab.get_engine().scenes)>0:
    mlab.close(all=True)
  mlab.figure(1)
  simlab.plotmesh(Th,legend=True)
  mlab.view(0,0)
  mlab.figure(2)
  simlab.plotmesh(Th,d=1,line_width=2,legend=True)
  simlab.plotmesh(Th,color='LightGray',opacity=0.05)
  mlab.view(0,0)
  mlab.figure(3)
  simlab.plot(Th,u)
  mlab.view(0,0)
  mlab.colorbar(title='u',orientation='vertical')
  mlab.figure(4)
  simlab.plotiso(Th,u,contours=15)
  mlab.view(0,0)
  mlab.colorbar(title='u',orientation='vertical')
  simlab.plotmesh(Th,color='LightGray',opacity=0.05)
  
if isMatplotlib():
  import matplotlib.pyplot as plt
  import fc_simesh_matplotlib.siMesh as siplt
  from fc_tools.matplotlib import set_axes_equal,DisplayFigures
  plt.close('all')
  plt.ion()
  DisplayFigures(nfig=4)
  plt.figure(1)
  siplt.plotmesh(Th,legend=True)
  siplt.plotmesh(Th,d=1,linewidths=2,color='Black')
  set_axes_equal()
  plt.figure(2)
  siplt.plotmesh(Th,d=1,linewidths=2,legend=True)
  siplt.plotmesh(Th,color='LightGray',alpha=0.05)
  set_axes_equal()
  plt.figure(3)
  siplt.plot(Th,u)
  plt.colorbar(label='u')
  plt.axis('off');set_axes_equal()
  plt.figure(4)
  siplt.plotiso(Th,u,contours=15)
  plt.colorbar(label='u')
  siplt.plotmesh(Th,color='LightGray',alpha=0.3)
  plt.axis('off');set_axes_equal()
  


from . import operators
#from . import FEM

import numpy as np
from scipy import linalg
from scipy import sparse
import itertools
import types
from math import *

def genericFunc(d,sFunc):
  S='np.vectorize(lambda x1'
  for i in range(2,d+1):
    S=S+',x'+str(i)
  S=S+':'+sFunc+')'
  #print(S)
  return eval(S)


def getVFindices(Num,m,nq):
  if (Num==1):
    return lambda I,i: I+i*nq
  else:
    return lambda I,i: m*I+i
  
def setFdata(f,Th,**kwargs):
  if f is None:
    return None
  dtype=kwargs.get('dtype', float)
  Num=kwargs.get('Num', 1)
  if np.isscalar(f):
    return f*np.ones((Th.nq,),dtype=dtype)
  if isinstance(f, type(lambda: None)):
    f=np.vectorize(f)
  if isinstance(f,np.lib.function_base.vectorize):
    #d=Th.q.shape[1]
    #S='f(Th.q[:,0]'
    #for i in range(1,d):
      #S=S+',Th.q[:,'+str(i)+']'
    #S=S+')'
    #V=eval(S)
    V=Th.feval(f)
    if np.isscalar(V):
      return V*np.ones((Th.nq,),dtype=dtype)
    return V
  if isinstance(f,np.ndarray) and (f.shape[0]==Th.nq) :
    return f
  if isinstance(f,list): # Vector Field case
    n=len(f)
    Fh=np.zeros((n*Th.nq,));I=np.arange(Th.nq)
    VFInd=getVFindices(Num,n,Th.nq)
    for i in range(n):
      Fh[VFInd(I,i)]=setFdata(f[i],Th)
    #Fh=np.zeros((n,Th.nq))
    #for i in range(n):
    #  Fh[i]=setFdata(f[i],Th,dtype=dtype)
    return Fh
  else:
    assert(0==1)

def ComputeGradientVec(q,me,vols):
  d=q.shape[1]
  if (d==1):
    return ComputeGradientVec1D(q,me,vols)
  elif (d==2):
    return ComputeGradientVec2D(q,me,vols)
  elif (d==3):
    return ComputeGradientVec3D(q,me,vols)
  else:
    return ComputeGradientVecdD(q,me,vols)

def ComputeGradientVec1D(q,me,vols):
  nme=me.shape[0]
  G=np.ndarray(shape=(2,1,nme))
  G[0,1]=1/vols
  G[0,0]=-G[0,1]
  
def ComputeGradientVec2D(q,me,vols):
  nme=me.shape[0]
  C=1/(2*vols)
  
  G=np.ndarray(shape=(3,2,nme))
  u=(q[me[:,1]]-q[me[:,2]]).T
  G[0,0]=u[1]*C
  G[0,1]=-u[0]*C
  u=(q[me[:,2]]-q[me[:,0]]).T
  G[1,0]=u[1]*C
  G[1,1]=-u[0]*C
  u=(q[me[:,0]]-q[me[:,1]]).T
  G[2,0]=u[1]*C
  G[2,1]=-u[0]*C
  return G
  
def ComputeGradientVec3D(q,me,vols):
  nme=me.shape[0]
  q1=q[me[:,0]].T;q2=q[me[:,1]].T;q3=q[me[:,2]].T;q4=q[me[:,3]].T
  C=1/(6*vols)
  D12=(q1-q2);D13=(q1-q3);D14=(q1-q4)
  D23=(q2-q3);D24=(q2-q4)
 
  G=np.ndarray(shape=(4,3,nme))
  G[0,0] = (D23[2]*D24[1] - D23[1]*D24[2])*C
  G[0,1] = (D23[0]*D24[2] - D23[2]*D24[0])*C
  G[0,2] = (D23[1]*D24[0] - D23[0]*D24[1])*C
  G[1,0] = (D13[1]*D14[2] - D13[2]*D14[1])*C
  G[1,1] = (D13[2]*D14[0] - D13[0]*D14[2])*C
  G[1,2] = (D13[0]*D14[1] - D13[1]*D14[0])*C
  G[2,0] = (D12[2]*D14[1] - D12[1]*D14[2])*C
  G[2,1] = (D12[0]*D14[2] - D12[2]*D14[0])*C
  G[2,2] = (D12[1]*D14[0] - D12[0]*D14[1])*C
  G[3,0] = (D12[1]*D13[2] - D12[2]*D13[1])*C
  G[3,1] = (D12[2]*D13[0] - D12[0]*D13[2])*C
  G[3,2] = (D12[0]*D13[1] - D12[1]*D13[0])*C
  return G
  
def ComputeGradientVecdD(q,me,vols):
  d=q.shape[1]
  nme=me.shape[0]
  Grad=np.zeros((d+1,d))
  Grad[1:d+1]=np.eye(d)
  Grad[0]=-1

  X=np.zeros((nme,d,d))
  for i in range(d):
    X[:,:,i]=q[me[:,i+1]]-q[me[:,0]]   
   
  #G=np.ndarray(shape=(d+1,d,nme))
  G1=np.array([np.dot(Grad,linalg.inv(X[k])) for k in range(nme)])
  #for k in range(nme):
    #G[::,::,k]=np.dot(Grad,linalg.inv(X[k]))
  return G1.swapaxes(0,1).swapaxes(1,2)


def NoneVector(d):
  V=[]
  for i in range(d):
    V.append(None)
  return V
  
def NoneMatrix(m,n):
  M=[]
  for i in range(m):
    M.append(NoneVector(n))
  return M
  
